#
# Copyright (c) 2002, Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# handle possibly compressed files
#

import time
import struct
import zlib
import sys
import os

class CompFile(file):
    """A class for transparently handling compressed files.

The CompFile class is a wrapper around the file object. It
is similar to Python's GzipFile class in gzip.py, but the
implementation takes advantage of features found in Python 2.2.

A CompFile object has the attributes and methods of a file
object, except a CompFile object cannot truncate a compressed
file.

It may be possible to add zipfile reading/writing capabilities
to the CompFile class in the future.
    """
    def __init__(self, filename, mode="r", buffering=0):
        if sys.platform == 'win32' and 'b' not in mode:
            mode += 'b'
        file.__init__(self, filename, mode, buffering)
        self.__size = 0
        self.__crc = zlib.crc32('')
        self.__compobj = None
        self.__offset = 0L
        self.__buffer = ''
        if mode == "r" or mode == "rb":
            magic = file.read(self, 2)
            if magic == '\037\213': # gzip/zlib file
                method = ord(file.read(self, 1))
                if method != 8:
                    raise IOError, "Unknown compression method"
                flags = ord(file.read(self, 1))
                data = file.read(self, 6) # skip mod time, extra flags, os
                if flags & 2: # file has FHCRC
                    self.__compobj = zlib.decompressobj()
                else:
                    self.__compobj = zlib.decompressobj(-zlib.MAX_WBITS)
                if flags & 4: # FEXTRA
                    xlen = ord(file.read(self,1))
                    xlen = xlen + 256*ord(file.read(self,1))
                    data = file.read(self, xlen)
                if flags & 8: # FNAME
                    while 1:
                        data = file.read(self, 1)
                        if not data or data == '\000':
                            break
                if flags & 16: # FCOMMENT
                    while 1:
                        data = file.read(self, 1)
                        if not data or data == '\000':
                            break
            else:
                file.seek(self, 0, 0)
        elif mode == "w" or mode == "wb":
            #
            # the following generates files that aren't
            # compatible with current gzip ...
            #
            # self.__compobj = zlib.compressobj()
            #
            # create a gzip-compatible compressobj ...
            self.__compobj = zlib.compressobj(6,
                                              zlib.DEFLATED,
                                              -zlib.MAX_WBITS,
                                              zlib.DEF_MEM_LEVEL,
                                              0)
            file.write(self, '\037\213\010')
            file.write(self, chr(8)) # flags FNAME - no FHCRC for gzip compat.
            file.write(self, struct.pack("<L", long(time.time())))
            file.write(self, '\000\377') # extra flags and unknown OS flag
            bname = os.path.basename(filename)
            if bname.endswith('.gz'):
                bname = bname[:-3]
            file.write(self, bname + '\000')

    def read(self, size=-1):
        if self.__compobj is None:
            data = file.read(self, size)
        elif size < 0:
            buf = file.read(self, size)
            self.__crc = zlib.crc32(buf, self.__crc)
            data = self.__buffer + self.__compobj.decompress(buf)
            data = data + self.__compobj.flush()
            self.__buffer = ''
        elif size < len(self.__buffer):
            data = self.__buffer[:size]
            self.__buffer = self.__buffer[size:]
        else:
            buf = file.read(self, size)
            while buf != '':
                self.__crc = zlib.crc32(buf, self.__crc)
                new_data = self.__compobj.decompress(buf)
                self.__buffer = self.__buffer + new_data
                if len(self.__buffer) > size:
                    break
                buf = file.read(self, size)
            self.__offset = len(self.__buffer)
            if self.__offset > size:
                data = self.__buffer[:size]
                self.__buffer = self.__buffer[size:]
            else:
                data = self.__buffer[:]
                self.__buffer = ''
        return data

    def readline(self, size=-1):
        if size > 0:
            data = self.read(size)
        elif self.__compobj is None:
            data = file.readline(self, size)
        elif self.__buffer.find("\n") != -1:
            idx = self.__buffer.find("\n") + 1
            data = self.__buffer[:idx]
            self.__buffer = self.__buffer[idx:]
        else:
            buf = self.read(100)
            offset = buf.find("\n")
            while offset < 0:
                new_buf = self.read(100)
                if new_buf == '': # EOF
                    break
                buf = buf + new_buf
                offset = buf.find("\n")
            if offset != -1:
                offset = offset + 1 # add in the newline character
                data = buf[:offset]
                self.__buffer = buf[offset:] + self.__buffer
            else:
                data = buf[:]
                self.__buffer = ''
        return data

    def readlines(self, size=0):
        if size <= 0:
            size = sys.maxint
        lines = []
        while size > 0:
            line = self.readline()
            if line == '':
                break
            lines.append(line)
            size = size - len(line)
        return lines
    
    def write(self, data):
        if len(data):
            self.__crc = zlib.crc32(data, self.__crc)
            self.__size = self.__size + len(data)
            file.write(self, self.__compobj.compress(data))

    def writelines(self, lines):
        for line in lines:
            self.write(line)

    def flush(self):
        if self.__compobj is not None:
            file.write(self, self.__compobj.flush())
        file.flush(self)
        
    def close(self):
        mode = self.mode
        if mode == "w" or mode == "wb":
            file.write(self, self.__compobj.flush())
            file.write(self, struct.pack("<l", self.__crc))
            file.write(self, struct.pack("<l", self.__size))
        file.close(self)

    def seek(self, offset, whence=0):
        if self.__compobj is None:
            file.seek(self, offset, whence)        
        elif self.mode == "r" or self.mode == "rb":
            if whence == 0: # from start
                file.seek(self, 0, 0)
                self.__buffer = ''
                magic = file.read(self, 2)
                if magic != '\037\213':
                    raise ValueError, "Invalid gzip file magic value!"
                method = ord(file.read(self, 1))
                if method != 8:
                    raise IOError, "Unknown compression method"
                flags = ord(file.read(self, 1))
                data = file.read(self, 6) # skip mod time, extra flags, os
                if flags & 2: # file has FHCRC
                    self.__compobj = zlib.decompressobj()
                else:
                    self.__compobj = zlib.decompressobj(-zlib.MAX_WBITS)
                if flags & 4: # FEXTRA
                    xlen = ord(file.read(self,1))
                    xlen = xlen + 256*ord(file.read(self,1))
                    data = file.read(self, xlen)
                if flags & 8: # FNAME
                    while 1:
                        data = file.read(self, 1)
                        if not data or data == '\000':
                            break
                if flags & 16: # FCOMMENT
                    while 1:
                        data = file.read(self, 1)
                        if not data or data == '\000':
                            break
                if offset > 0:
                    data = self.read(offset)
            elif whence == 1: # current position
                if offset < 0:
                    self.seek(0,0)
                    new_offset = self.__offset + offset
                    if new_offset > 0:
                        self.read(new_offset)
                elif offset > 0:
                    self.read(offset)
            elif whence == 2: # from end
                buf = ''
                self.seek(0,0)
                data = self.read(4096)
                while data != '':
                    buf = buf + data
                    if len(buf) > offset:
                        idx = 1 - offset
                        buf = buf[-idx:]
                    data = self.read(4096)
                if len(buf) > offset:
                    index = len(buf) - offset
                    self.__buffer = buf[offset:]
                else:
                    raise IOError, "Offset larger than filesize: " + `offset`
            else:
                raise ValueError, "Invalid seek position: " + `whence`
        else:
            raise IOError, "Unable to seek on writing."

    def truncate(self, size=None):
        if self.__compobj is not None:
            raise StandardError, "Cannot truncate compressed files."
        if size is None:
            size = file.tell(self)
        file.truncate(self, size)



#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# menu code
#
# the menu things below should remain interface neutral
#

primary_menus = [('_File', 'FILE_MENU', 'file_menu_init'),
                 ('_Edit', 'EDIT_MENU', 'edit_menu_init'),
                 ('_Draw', 'DRAW_MENU', 'draw_menu_init'),
                 ('_Modify', 'MODIFY_MENU', 'modify_menu_init'),
                 ('D_imensions', 'DIMENSION_MENU', 'dimension_menu_init'),
                 ('De_bug', 'DEBUG_MENU', 'debug_menu_init'),
                 ]
submenus = {}

file_menu = [ ['_New', 'file_new', 'NEW', 'N' ],
              ['_Open', 'file_open', 'OPEN', 'O'],
              ['_Close', 'file_close', 'CLOSE', 'W'],
              [None],
              ['_Save', 'file_save', 'SAVE', 'S'],
              ['Save As ...', 'file_save_as', 'SAVE_AS', None],
              ['Save Layer As ...', 'file_save_layer', None, None],
              [None],
              ['_Quit', 'file_quit', 'QUIT', 'Q'],
              ]

submenus['FILE_MENU'] = file_menu

edit_menu = [ ['Undo', 'edit_undo', 'UNDO', 'Z'],
              ['Redo', 'edit_redo', 'REDO', None],
              [None],
              ['Cut', 'edit_cut', 'CUT', 'X'],
              ['Copy', 'edit_copy', 'COPY', 'C'],
              ['Paste', 'edit_paste', 'PASTE', 'V'],
              [None],
              ['Select', 'select_cb', None, None],
              ['Select All ...', 'EDIT_SELECT_ALL_MENU', 'select_all_init'],
              [None],
              ['Preferences', 'prefs', 'PREFS', None]
              ]

submenus['EDIT_MENU'] = edit_menu

edit_select_all_menu = [ ['Points', 'select_all_points', None, None],
                         ['Segments', 'select_all_segments', None, None],
                         ['Circles', 'select_all_circles', None, None],
                         ['Arcs', 'select_all_arcs', None, None],
                         ['HCLines', 'select_all_hclines', None, None],
                         ['VCLines', 'select_all_vclines', None, None],
                         ['ACLines', 'select_all_aclines', None, None],
                         ['CLines', 'select_all_clines', None, None],
                         ['CCircles', 'select_all_ccircles', None, None],
                         ['Linear Dims', 'select_all_ldims', None, None],
                         ['Horiz. Dims', 'select_all_hdims', None, None],
                         ['Vert. Dims', 'select_all_vdims', None, None],
                         ['Radial Dims.', 'select_all_rdims', None, None],
                         ['Angular Dims.', 'select_all_adims', None, None],
                         ]

submenus['EDIT_SELECT_ALL_MENU'] = edit_select_all_menu

draw_menu = [ ['Basic', 'DRAW_BASIC_MENU', 'draw_basic_init'],
              ['Con. Lines', 'DRAW_CONLINES_MENU', 'draw_conlines_init'],
              ['Con. Circles', 'DRAW_CONCIRCS_MENU', 'draw_concircs_init'],
              ['Chamfer', 'draw_chamfer', None, None],
              ['Fillet', 'draw_fillet', None, None],
              [None],
              ['Leader', 'draw_leader', None, None],
              [None],
              ['Polyline', 'draw_polyline', None, None],
              [None],
              ['Polygon', 'draw_polygon', None, None],
              ['Polygon (ext)', 'draw_ext_polygon', None, None],
              [None],
              ['Text', 'draw_text', None, None],
              [None],
              ['Set ...', 'DRAW_SET_MENU', 'draw_set_init'],
              [None],
              ['Add New ...', 'DRAW_ADD_MENU', 'draw_add_init']
              ]

submenus['DRAW_MENU'] = draw_menu

draw_basic_menu = [ ['Points', 'draw_point' , None, None],
                    ['Segments', 'draw_segment', None, None],
                    ['Rectangles', 'draw_rectangle', None, None],
                    ['Circle (cen)', 'draw_circle_center', None, None],
                    ['Circle (2 pts)', 'draw_circle_tp', None, None],
                    ['Arc', 'draw_arc_center', None, None],
                    ]

submenus['DRAW_BASIC_MENU'] = draw_basic_menu

draw_conlines_menu = [ ['Horizontal', 'draw_hcl', None, None],
                       ['Vertical', 'draw_vcl', None, None],
                       ['Angled', 'draw_acl', None, None],
                       ['Two-Point', 'draw_cl', None, None],
                       ['Perpendicular', 'draw_perpendicular_cline', None, None],
                       ['Tangent', 'draw_tangent_cline', None, None],
                       ['Tangent 2 Circs', 'draw_tangent_two_ccircles', None, None],
                       ['Parallel', 'draw_poffset_cline', None, None],
                       ]

submenus['DRAW_CONLINES_MENU'] = draw_conlines_menu

draw_concircs_menu = [ ['Center Point', 'draw_ccirc_cp', None, None],
                       ['Two-Point', 'draw_ccirc_tp', None, None],
                       ['Tangent', 'DRAW_TANGENT_CCIRCLE_MENU', 'draw_tangent_ccircle_menu_init']
                       ]

submenus['DRAW_CONCIRCS_MENU'] = draw_concircs_menu

draw_tangent_ccircle_menu = [ ['One object', 'draw_tangent_single_conobj', None, None],
                              ['Two objects', 'draw_tangent_two_conobjs', None, None],
                              ]

submenus['DRAW_TANGENT_CCIRCLE_MENU'] = draw_tangent_ccircle_menu

draw_set_menu = [ ['Style', 'draw_set_style', None, None],
                  ['Linetype', 'draw_set_linetype', None, None],
                  ['Color', 'draw_set_color', None, None],
                  ['Thickness', 'draw_set_thickness', None, None],
                  ]

submenus['DRAW_SET_MENU'] = draw_set_menu

draw_add_menu = [ ['Style', 'cb', None, None],
                  ['Linetype', 'cb', None, None],
                  ['Color', 'cb', None, None],
                  ]

submenus['DRAW_ADD_MENU'] = draw_add_menu

modify_menu =  [ ['Move ...', 'MODIFY_MOVE_MENU', 'modify_move_init'],
                 ['Stretch ...', 'MODIFY_STRETCH_MENU', 'modify_stretch_init'],
                 ['Split', 'split', None, None],
                 ['Mirror', 'mirror', None, None],
                 ['Transfer', 'transfer', None, None],
                 [None],
                 ['Delete', 'delete', 'DELETE', None],
                 [None],
                 ['Change ...', 'MODIFY_CHANGE_MENU', 'modify_change_init'],
                 [None],
                 ['Zoom', 'zoom', None, None],
                 ['Zoom In', 'zoom_in', 'ZOOM_IN', None],
                 ['Zoom Out', 'zoom_out', 'ZOOM_OUT', None],
                 ['Zoom Fit', 'zoom_fit', 'ZOOM_FIT', None],
                 ]
submenus['MODIFY_MENU'] = modify_menu

modify_move_menu = [ ['Horizontal' ,'move_horiz', None, None],
                     ['Vertical', 'move_vert', None, None],
                     ['Both', 'move_twopoint', None, None],
                     ]

submenus['MODIFY_MOVE_MENU'] = modify_move_menu

modify_stretch_menu = [ ['Horizontal', 'stretch_horiz', None, None],
                        ['Vertical', 'stretch_vert', None, None],
                        ['Both', 'stretch_twopoint', None, None],
                        ]

submenus['MODIFY_STRETCH_MENU'] = modify_stretch_menu

modify_change_menu = [ ['Style', 'change_style', None, None],
                       ['Linetype', 'change_linetype', None, None],
                       ['Color', 'change_color', None, None],
                       ['Thickness', 'change_thickness', None, None],
                       ]

submenus['MODIFY_CHANGE_MENU'] = modify_change_menu

dimension_menu = [ ['Linear', 'dimension_linear' , None, None ],
                   ['Horizontal', 'dimension_horiz', None, None ],
                   ['Vertical', 'dimension_vert', None, None ],
                   ['Radial', 'dimension_rad', None, None],
                   ['Angular', 'dimension_ang', None, None ],
                   ]

submenus['DIMENSION_MENU'] = dimension_menu

debug_menu = [ ['Focus', 'get_focus_widget', None, None],
               ['UndoStack', 'undo_stack', None, None],
               ['RedoStack', 'redo_stack', None, None],
               ['ImageUndo', 'image_undo', None, None],
               ['ImageRedo', 'image_redo', None, None],
               ]

submenus['DEBUG_MENU'] = debug_menu



#
# Copyright (c) 2002-2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Cocoa CAD drawing view 
#

from objc import nil

from PyObjCTools import NibClassBuilder
from Foundation import *
from AppKit import *

from math import hypot, ceil
import Generic.graphicobject
import Generic.layer
import Generic.globals
import Generic.segment
import Generic.circle
import Generic.arc
import Generic.hcline
import Generic.vcline
import Generic.acline
import Generic.cline
import Generic.ccircle
import Generic.conobject
import Generic.segjoint
import Generic.leader
import Generic.polyline
import Generic.text
import Generic.dimension
import Generic.tools

import Interface.Cocoa.Globals
import Interface.Cocoa.ImageDocument
import Interface.Cocoa.ImageWindowController
import Interface.Cocoa.AppController


#
# Define True & False in case this is Python < 2.2.1
#
try:
    True, False
except NameError:
    True, False = 1, 0
    
#
# Global variables
#
Generic.globals.NSColors = {}
Generic.globals.NSFonts = {}

NibClassBuilder.extractClasses("ImageDocument")


class CADView(NibClassBuilder.AutoBaseClass, Generic.message.Messenger):
    """ Custom NSView for visual display & editing of a PythonCad CAD image.
    
A CADView is an NSView used for editing & display of a generic Image object.

    """

#
# Beginning of Cocoa methods
#
    def init(self):
        """ NSView override for cocoa initializer of CADView
        
init()
        """
        self = super(CADView, self).init()
        if (self):
            self.__mouseTrackTimer = None
            self.__document = None
            self.__tracking_rect = None
            self.__trans = None
            self.__x_scale = 1.0
            self.__y_scale = 1.0
            
        return self
            

    def awakeFromNib(self):
        """ NSView override for post-NIB loading initialization

awakeFromNib(self)
        """
        #
        # initialize variables
        #
        self.__mouseTrackTimer = None
        self.__document = None
        self.__tracking_rect = None
        self.__trans = None
        self.__temporaryObject = []
        self.__temporaryPoint = NSMakePoint(0.0,0.0)
        self.__temporaryRect = NSMakeRect(0.0, 0.0, 0.0, 0.0)
        self.__x_scale = 1.0
        self.__y_scale = 1.0
        self.updateTrackingRect()
        self.setTransform()
        _cv = self.enclosingScrollView().contentView()
        #
        # register for frame updates
        #
        _nc = NSNotificationCenter.defaultCenter()
        _nc.addObserver_selector_name_object_(self, "frameChangeNotification:", "NSViewFrameDidChangeNotification", _cv)
        _nc.addObserver_selector_name_object_(self, "boundsChangeNotification:", "NSViewBoundsDidChangeNotification", self)
        _nc.addObserver_selector_name_object_(self, "windowCloseNotification:", "NSWindowWillCloseNotification", self.window())
                
    def isOpaque(self):
        _bool = self.inLiveResize()
        if _bool:
            return False
        return True
        
    def getDocument(self):
        """ Gets ImageDocument displayed by this view
        
getDocument()
        """
        if self.__document is None:
            self.__document = self.window().windowController().document()
        return self.__document
                        
    def setMouseTimer(self, timer):
        """ Sets timer for tracking mouse location
        
setMouseTimer(timer)
        """
        if (self.__mouseTrackTimer) is not None:
            self.__mouseTrackTimer.invalidate()
        self.__mouseTrackTimer = timer
        
    def setTrackingRect(self, rect):
        """ Sets up tracking rectange for mouse location
        
setTrackingRect(rect)
        """
        if (self.__tracking_rect) is not None:
            self.removeTrackingRect_(self.__tracking_rect)
        self.__tracking_rect = rect
          
                    
        
    def windowCloseNotification_(self, note):
        _nc = NSNotificationCenter.defaultCenter()
        _nc.removeObserver_(self)
        self.setMouseTimer(None)
        
    def frameChangeNotification_(self, note):
        """ NSView message for changes in enclosing clip view frame size
        
frameChangeNotification(note)
        """
        #
        # if the area for the CADView is bigger than the CADview,
        # we make the CADView bigger.  It might make sense later to
        # shrink the CADView IF we can do so without hiding any images
        #
        _cv = note.object()
        _cvsize = _cv.frame().size
        _mysize = self.frame().size
        _redraw = False
        if (_cvsize.width > _mysize.width):
            _mysize.width = _cvsize.width
            _redraw = True
        if (_cvsize.height > _mysize.height):
            _mysize.height = _cvsize.height
            _redraw = True
        if (_redraw):
            self.setFrameSize_(_mysize)
            self.setNeedsDisplay_(True)
        self.updateTrackingRect()
        Generic.globals.NSFonts.clear()
        self.setTransform()


    def boundsChangeNotification_(self, note):
        """ NSView message for changes in bounds size
        
boundsChangeNotification(note)
        """
        self.setTransform()
        Generic.globals.NSFonts.clear()
        
    def itemAddedNotification_(self, note):
        """ ImageDocument message for new item added to drawing

itemAddedNotification(note)
        """
        self.setNeedsDisplayInRect_(self._getTempRect())
        self._clearTempItems()
        _info = note.userInfo()
        if not _info.has_key("layer"):
            return
        if not _info["layer"].isVisible():
            return
        if _info.has_key("item") and self.canDraw():
            _item = _info["item"]
            _objList = [_item]
            # 
            # assume item added to active layer - might be a problem
            #
            self.lockFocus()
            _bounds = self.bounds()
            if isinstance(_item, Generic.conobject.ConstructionObject):
                self._drawConstructionObjects(_objList, _bounds, True)
                
            elif isinstance(_item, Generic.point.Point):
                if self.getDocument().getOption('HIGHLIGHT_POINTS'):
                    self._drawPoints(_objList, _bounds)
            else:
                self._drawObjects(_objList, _bounds, True)
                
            self.unlockFocus()

    def itemDeletedNotification_(self, note):
        """ ImageDocument message for item deleted from drawing

itemDeletedNotification(note)
        """
        _info = note.userInfo()
        if not _info.has_key("layer"):
            return
        if not _info["layer"].isVisible():
            return
        self.setNeedsDisplay_(True)
            
    def setTempPoint(self, point=NSMakePoint(0.0, 0.0)):
        self.__temporaryPoint = point
            
    def setTempObject(self, obj=None, flush=True):
        if obj is None:
            self.__temporaryObject = []
        elif isinstance(obj, (Generic.graphicobject.GraphicObject,
                              Generic.graphicobject.old_GraphicObject,
                              Generic.text.TextBlock,
                              Generic.dimension.Dimension)):
                if flush:
                    self.__temporaryObject=[obj]
                else:
                    self.__temporaryObject.append(obj)
        else:
            raise TypeError, "Invalid graphics object:" + `obj`

    def flushTempObjects(self):
        self.setNeedsDisplayInRect_(self._getTempRect())
        self._clearTempItems()

    def _setTempRect(self, rect=NSMakeRect(0.0,0.0,0.0,0.0)):
        self.__temporaryRect = rect
        
    def _getTempRect(self):
        return self.__temporaryRect

    def _getTempObject(self):
        return self.__temporaryObject
        
    def _clearTempItems(self):
        self.setTempPoint()
        self._setTempRect()
        self.setTempObject(None)
            
    def getImageRep(self):
        _frame = self.frame() 
        _img = NSImage.alloc().initWithSize_(_frame.size)
        _img.setScalesWhenResized_(False)
        _img.lockFocus()
        self.drawRect_(self.frame())
        _img.unlockFocus()
        return _img
          
#    def viewWillStartLiveResize(self):
#        super(CADView, self).viewWillStartLiveResize()
#        _bounds = self.bounds() 
#        _image = NSImage.alloc().initWithSize_(_bounds.size) 
#        _image.setCachedSeparately_(True)
#        _image.lockFocus()
#        self.drawRect_(_bounds)
#        _image.unlockFocus()
#        self.__image = NSImageView.alloc().initWithFrame_(_bounds)
#        self.addSubview_(self.__image)
#        self.__image.setAutoresizingMask_(NSViewWidthSizable|NSViewHeightSizable)
#        self.__image.setBounds_(_bounds)
#        self.__image.setImage_(_image)
        
       
    def postLoadSetup(self):
        """ Called after window loaded to make sure things register correctly.
        Should be able to do this with built-in functions, but can't get it to work.
        """
        _img = self.getDocument().getImage()
        _obj = Interface.Cocoa.Globals.wrap(_img)
        _nc = NSNotificationCenter.defaultCenter()
        _nc.addObserver_selector_name_object_(self,"itemAddedNotification:","added_object", _obj)
        _nc.addObserver_selector_name_object_(self,"itemDeletedNotification:","deleted_object", _obj)
     

    def viewDidEndLiveResize(self):
        super(CADView, self).viewDidEndLiveResize()
#        self.__image.removeFromSuperviewWithoutNeedingDisplay()
#        del self.__image
#        self.__image = None
#        self.setNeedsDisplay_(True)
        self.updateTrackingRect()
         
    def updateTrackingRect(self):
        """ Sets tracking rectangle over frame boundary to track mouse
        
updateTrackingRect()
        """
        _point = self.getMouseLocation()
        _vis = self.visibleRect()
        _flag = self.mouse_inRect_(_point, _vis)
        _rect = self.addTrackingRect_owner_userData_assumeInside_(_vis, self, 0, _flag)
        self.setTrackingRect(_rect)
        
    def mouseEntered_(self, event):
        """ NSView override to announce mouse entered CADView.  Use for location update.
        
mouseEntered_(event)
        """
        # fire timer every 0.1 sec to track mouse moves
        
        _timer = NSTimer.scheduledTimerWithTimeInterval_target_selector_userInfo_repeats_(0.1, self, "updateMouseLocation:", 0, True)
        self.setMouseTimer(_timer)
    
    def mouseExited_(self, event):
        """ NSView override to announce mouse exited CADView.  Use for location update.

mouseExited_(event)
        """
        self.setMouseTimer(None)
        
    def mouseDown_(self, event):
        _doc = self.getDocument()
        _tool = _doc.getTool()
        if _tool is not None and _tool.hasHandler("left_button_press"):
            _handler = _tool.getHandler("left_button_press")
            _handler(_doc, event, _tool)

    def updateMouseLocation_(self, timer):
        """ timer callback that updates current mouse location in ImageDocument
        
updateMouseLocation_(timer):
        """
        _point = self.getMouseLocation()
        _doc = self.getDocument()
        _ox, _oy = _doc.getPoint()
        if (_ox == _point.x) and (_oy == _point.y):
            return
        _doc.setPoint(_point.x,_point.y)
        _tool = _doc.getTool()
        if _tool is not None and _tool.hasHandler("mouse_move"):
            _handler = _tool.getHandler("mouse_move")
            _handler(_doc, _point, _tool)
            _x = _y = _w = _h = None
            
            if isinstance(_tool, (Generic.tools.TwoPointCircleTool,
                                  Generic.tools.TangentCCircleTool,
                                  Generic.tools.TwoPointTangentCCircleTool)):
                _r = _tool.getRadius()
                if _r is not None and _r > 0:
                    _cx, _cy = _tool.getCenter()
                    _w = _h = _r * 2.1
                    _x = _cx - _r * 1.05
                    _y = _cy - _r * 1.05
                else:
                    return
                    
            elif isinstance(_tool, Generic.tools.ArcTool):
                _r = _tool.getRadius()
                if _r is not None and _r > 0:
                    _cx, _cy = _tool.getCenter()
                    _w = _h = _r * 2.1
                    _x = _cx - _r * 1.05
                    _y = _cy - _r * 1.05
                              
            elif isinstance(_tool, Generic.tools.CircleTool):
                _r = hypot((self.__temporaryPoint.x - _point.x), (self.__temporaryPoint.y - _point.y))
                _w = _h = _r * 2.1
                _x = self.__temporaryPoint.x - _r * 1.05
                _y = self.__temporaryPoint.y - _r * 1.05
                
            elif isinstance(_tool, Generic.tools.CCircleTangentLineTool):
                _tanpts = _tool.getTangentPoints()
                _tanrect = NSMakeRect(0,0,0,0)
                for _set in _tanpts:
                    _x1, _y1, _x2, _y2 = _set
                    _wt = abs(_x2 - _x1) * 1.06
                    _ht = abs(_y2 - _y1) * 1.06
                    _xt = min(_x1, _x2) - _wt * 0.03
                    _yt = min(_y1, _y2) - _ht * 0.03
                    _setrect = NSMakeRect(_xt, _yt, _wt, _ht)
                    _tanrect = NSUnionRect(_tanrect, _setrect)
                _x = NSMinX(_tanrect)
                _y = NSMinY(_tanrect)
                _w = NSWidth(_tanrect)
                _h = NSHeight(_tanrect) 
                
            elif isinstance(_tool, Generic.tools.PolygonTool):
                _count = _tool.getSideCount()
                (_xc, _yc) = _tool.getCoord(0)
                _x = _xmax = _xc
                _y = _ymax = _yc
                for _i in range(1, _count):
                    (_xc, _yc) = _tool.getCoord(_i)
                    _x = min(_x, _xc)
                    _xmax = max(_xmax, _xc)
                    _y = min(_y, _yc)
                    _ymax = max(_ymax, _yc)
                _w = abs(_xmax - _x) * 1.06
                _h = abs(_ymax - _y) * 1.06
                _x = _x - _w*0.03
                _y = _y - _h*0.03
                
            elif isinstance(_tool, Generic.tools.TextTool):
                _x = _point.x
                _y = _point.y
                _vis = self.visibleRect()
                _w = NSWidth(_vis)/10.0
                _h = NSHeight(_vis)/10.0

            elif isinstance(_tool, Generic.tools.AngularDimensionTool):
                _l1, _fp = _tool.getFirstPoint()
                _l2, _sp = _tool.getSecondPoint()
                if (_fp is not None) and (_sp is not None):
                    _l, _vp = _tool.getVertexPoint()
                    _vx, _vy = _vp.getCoords()
                    _rad = hypot((_vx - _point.x), (_vy - _point.y))
                    _dim = _tool.getDimension()
                    _bar1, _bar2 = _dim.getDimBars()
                    _ep1, _ep2 = _bar1.getEndpoints()
                    _ep3, _ep4 = _bar2.getEndpoints()
                    _x1, _y1 = _ep1
                    _x2, _y2 = _ep2
                    _x3, _y3 = _ep3
                    _x4, _y4 = _ep4
                    _x5 = _vx - _rad
                    _x6 = _vx + _rad
                    _y5 = _vy - _rad
                    _y6 = _vy + _rad
                    _xmax = max(_x1, _x2, _x3, _x4, _x5, _x6, _point.x) 
                    _ymax = max(_y1, _y2, _y3, _y4, _y5, _y6, _point.y) 
                    _xmin = min(_x1, _x2, _x3, _x4, _x5, _x6, _point.x) 
                    _ymin = min(_y1, _y2, _y3, _y4, _y5, _y6, _point.y)
                    _w = abs(_xmax - _xmin) * 1.1
                    _h = abs(_ymax - _ymin) * 1.1
                    _x = _xmin - _w * 0.05
                    _y = _ymin - _h * 0.05
                    
            
            elif isinstance(_tool, Generic.tools.LinearDimensionTool):
                _l1, _p1 = _tool.getFirstPoint()
                _l2, _p2 = _tool.getSecondPoint()
                if _p1 is not None and _p2 is not None:
                    _dim = _tool.getDimension()
                    _bar1, _bar2 = _dim.getDimBars()
                    _crossbar = _dim.getDimCrossbar()
                    #
                    # draw dimension lines
                    #
                    _ep1, _ep2 = _bar1.getEndpoints()
                    _ep3, _ep4 = _bar2.getEndpoints()
                    _ep5, _ep6 = _crossbar.getEndpoints()
                    _x1, _y1 = _ep1
                    _x2, _y2 = _ep2
                    _x3, _y3 = _ep3
                    _x4, _y4 = _ep4
                    _x5, _y5 = _ep5
                    _x6, _y6 = _ep6
                    _xmax = max(_x1, _x2, _x3, _x4, _x5, _x6, _point.x) 
                    _ymax = max(_y1, _y2, _y3, _y4, _y5, _y6, _point.y) 
                    _xmin = min(_x1, _x2, _x3, _x4, _x5, _x6, _point.x) 
                    _ymin = min(_y1, _y2, _y3, _y4, _y5, _y6, _point.y)
                    _w = abs(_xmax - _xmin) * 1.1
                    _h = abs(_ymax - _ymin) * 1.1
                    if (_w < 50):
                        _w = 50
                    if (_h < 30):
                        _h = 30
                    _x = _xmin - _w * 0.05
                    _y = _ymin - _h * 0.05
                    
            elif isinstance(_tool, Generic.tools.RadialDimensionTool):
                _rdim = _tool.getDimension()
                _cb = _rdim.getDimCrossbar()
                _ep1, _ep2 = _cb.getEndpoints()
                _x1, _y1 = _ep1
                _x2, _y2 = _ep2
                _xmax = max(_x1, _x2, _point.x)
                _ymax = max(_y1, _y2, _point.y)
                _xmin = min(_x1, _x2, _point.x)
                _ymin = min(_y1, _y2, _point.y)
                _w = abs(_xmax - _xmin) * 1.2
                _h = abs(_ymax - _ymin) * 1.2
                if (_w < 50):
                    _w = 50
                if (_h < 30):
                    _h = 30
                _x = _xmin - _w * 0.15
                _y = _ymin - _h * 0.15
             
                   
                
            if (_x is None) or (_y is None) or (_w is None) or (_h is None): 
                _w = abs(_point.x - self.__temporaryPoint.x) * 1.06
                _h = abs(_point.y - self.__temporaryPoint.y) * 1.06
                _x = min(_point.x, self.__temporaryPoint.x) - _w * 0.03
                _y = min(_point.y, self.__temporaryPoint.y) - _h * 0.03
            _rect = NSMakeRect(_x, _y, _w, _h)
            _trect = self._getTempRect()
            self.setNeedsDisplayInRect_(_trect)
            self.setNeedsDisplayInRect_(_rect)
            self._setTempRect(_rect)
 
    def getMouseLocation(self):
        """ Finds mouse location in view
        
getMouseLocation()

This function returns an NSPoint corresponding to
mouse location in view's coordinate system.
        """
        _mouseLoc = self.window().mouseLocationOutsideOfEventStream()
        return self.convertPoint_fromView_(_mouseLoc, None)

    def fitImage(self):
        """Redraw the image so all entities are visible in the window.

fitImage()        
        """
        _clipView = self.enclosingScrollView().contentView()
        _conRect = _clipView.documentRect()
        _doc = self.getDocument()
        _xmin, _ymin, _xmax, _ymax = _doc.getImage().getExtents()
        _xdiff = abs(_xmax - _xmin) * 1.05
        _ydiff = abs(_ymax - _ymin) * 1.05
        _xs = _xdiff / NSWidth(_conRect) 
        _ys = _ydiff / NSHeight(_conRect)
        _xmid = (_xmin + _xmax) * 0.5
        _ymid = (_ymin + _ymax) * 0.5
        if _xs > _ys:
            _width = _xdiff
            _height = NSHeight(_conRect) * _xs
        else:
            _height = _ydiff
            _width = NSWidth(_conRect) * _ys
        _nx = _xmid - _width * 0.5
        _ny = _ymid - _height * 0.5
        self.zoomToRect(((_nx, _ny), (_width, _height)))
        
    def zoomToRect(self, rect):
        _clipView = self.enclosingScrollView().contentView()
        (_nx, _ny), (_nw, _nh) = rect
        (_cx, _cy), (_cw, _ch) = _clipView.frame()
        (_fx, _fy), (_fw, _fh) = self.frame()
        (_vx, _vy), (_vw, _vh) = self.visibleRect()
        _wzoom = _vw / _nw
        _hzoom = _vh / _nh
        _zoom = min(_wzoom, _hzoom)
        _nfw = _fw * _zoom
        _nfh = _fh * _zoom
        if _nfw < _cw or _nfh < _ch:
            _nfw = _cw
            _nfh = _ch
        self.setFrameSize_((_nfw, _nfh))
        (_bx, _by), (_bw, _bh) = self.bounds()
        self.setBoundsSize_((_bw / _zoom, _bh / _zoom))
        self.scrollRectToVisible_(rect)
        self.setNeedsDisplay_(True)
  
        
    def drawRect_(self, rect):
        """ NSView override to draw image in view
        
drawRect(rect)

Drawing view clipped to passed in NSRect "rect"
        """
        _doc = self.getDocument()
        _gc = NSGraphicsContext.currentContext()
        _gc.setShouldAntialias_(False)            
        self._drawBackground(_doc, rect)
        _active_layer = _doc.getImage().getActiveLayer()
        _layers = []
        _layers.append(_doc.getImage().getTopLayer())
        while (len(_layers)):
            _layer = _layers.pop()
            if _layer is not _active_layer and _layer.isVisible():
                self._drawLayer(_doc, _layer, rect)
            if _layer.hasSublayers():
                _layers.extend(_layer.getSublayers())
        if _active_layer.isVisible():
            self._drawLayer(_doc, _active_layer, rect)
            
    def _drawBackground(self, doc, rect):
        """ draws background color of ImageDocument doc in NSRect rect
        
_drawBackground(doc, rect)
        """
        _color = doc.getOption('BACKGROUND_COLOR')
        _nscolor = self.getNSColor(_color)
        _nscolor.set()
        NSRectFill(rect)
        
    def _drawLayer(self, doc, layer, rect):
        """ draws Layer "layer" of ImageDocument "doc" in NSRect "rect"
        
_drawLayer(doc, layer, rect)
        """
        _size = self.pointSize()
        _bigRect = NSInsetRect(rect, -_size.width, -_size.height)
        (_xmin, _ymin, _xmax, _ymax) = self.rectToCoordTransform(_bigRect)
        _objs = layer.objsInRegion(_xmin, _ymin, _xmax, _ymax)
        _active_layer = doc.getImage().getActiveLayer()
        if _active_layer is layer:
            active = True
            _tempObj = self._getTempObject()
            if len(_tempObj) > 0:
                for _obj in _tempObj:
                    _objs.append(_obj)
        else:
            active = False

        _conobjs = []
        _vobjs = []
        _pts = []

        for _obj in _objs:
            if isinstance(_obj, Generic.conobject.ConstructionObject):
                _conobjs.append(_obj)
            elif isinstance(_obj, Generic.point.Point):
                _pts.append(_obj)
            else:
                _vobjs.append(_obj)
        
        _bounds = self.bounds()
        self._drawConstructionObjects(_conobjs, _bounds, active)
        if doc.getOption('HIGHLIGHT_POINTS'):
            self._drawPoints(_pts, _bounds)
        self._drawObjects(_vobjs, _bounds, active)
                
        
    def _drawConstructionObjects(self, objs, rect, active):
        #
        # style bits are the same for all construction objects, so all
        # in the same bezier path
        #
        _constyle = Generic.conobject.ConstructionObject._ConstructionObject__defstyle
        _dashArray = _constyle.getLinetype().getList()
        if _dashArray is None:
            _dashArray = [2, 2]
        _scaledArray = []
        for _num in _dashArray:
            _scaledArray.append(_num*self.__x_scale)
        _path = NSBezierPath.bezierPath()
        _path.setLineDash_count_phase_(_scaledArray, len(_scaledArray), 0.0)
        _path.setLineCapStyle_(NSButtLineCapStyle)
        _path.setLineJoinStyle_(NSMiterLineJoinStyle)
        _path.setLineWidth_(0.0)
        if active:
            _color = _constyle.getColor()
        else:
            _color = self.getDocument().getOption('INACTIVE_LAYER_COLOR')
        self.getNSColor(_color).set()
        (_xmin, _ymin, _xmax, _ymax) = self.rectToCoordTransform(rect)
        for _obj in objs:
            if isinstance(_obj, Generic.hcline.HCLine):
                _x, _y = _obj.getLocation().getCoords()
                _tp = self.nudgePoint(rect, NSMakePoint(_x, _y))
                _point = NSMakePoint(_xmin, _tp.y)
                _path.moveToPoint_(_point)
                _point = NSMakePoint(_xmax, _tp.y)
                _path.lineToPoint_(_point)
            elif isinstance(_obj, Generic.vcline.VCLine):
                _x, _y = _obj.getLocation().getCoords()
                _tp = self.nudgePoint(rect, NSMakePoint(_x, _y))
                _point = NSMakePoint(_tp.x, _ymin)
                _path.moveToPoint_(_point)
                _point = NSMakePoint(_tp.x, _ymax)
                _path.lineToPoint_(_point)
            elif isinstance(_obj, Generic.acline.ACLine):
                _coords = _obj.clipToRegion(_xmin, _ymin, _xmax, _ymax)
                if _coords is not None:
                    _x1, _y1, _x2, _y2 = _coords
                    _tp1 = self.nudgePoint(rect, NSMakePoint(_x1, _y1))
                    _tp2 = self.nudgePoint(rect, NSMakePoint(_x2, _y2))
                    _path.moveToPoint_(_tp1)
                    _path.lineToPoint_(_tp2)
            elif isinstance(_obj, Generic.cline.CLine):
                _coords = _obj.clipToRegion(_xmin, _ymin, _xmax, _ymax)
                if _coords is not None:
                    _x1, _y1, _x2, _y2 = _coords
                    _tp1 = self.nudgePoint(rect, NSMakePoint(_x1, _y1))
                    _tp2 = self.nudgePoint(rect, NSMakePoint(_x2, _y2))
                    _path.moveToPoint_(_tp1)
                    _path.lineToPoint_(_tp2)
            elif isinstance(_obj, Generic.ccircle.CCircle):
                _x, _y = _obj.getCenter().getCoords()
                _radius = _obj.getRadius()
                _tp = self.nudgePoint(rect, NSMakePoint(_x - _radius, _y - _radius))
                _rect = NSMakeRect(_tp.x, _tp.y, _radius*2.0, _radius*2.0)
                _path.appendBezierPathWithOvalInRect_(_rect)
            else:
                pass
        _path.stroke()
       
    def _drawPoints(self, pts, rect):
        _ptSize = self.pointSize()
        _xcorr = _ptSize.width * 0.5
        _ycorr = _ptSize.height * 0.5
        _pathSingle = NSBezierPath.bezierPath()
        _pathSingle.setLineCapStyle_(NSButtLineCapStyle)
        _pathSingle.setLineJoinStyle_(NSMiterLineJoinStyle)
        _pathSingle.setLineWidth_(0.0)
        _pathMulti = _pathSingle.copyWithZone_(None)
        for _pt in pts:
            _users = _pt.getUsers()
            _path = _pathSingle
            if len(_users) > 1:
                _nd = 0
                for _uref in _users:
                    _user = _uref()
                    if not isinstance(_user, Generic.dimension.Dimension):
                        _nd = _nd + 1
                    if _nd > 1:
                        _path = _pathMulti
                        break
            _orig = NSMakePoint(_pt.x - _xcorr, _pt.y - _ycorr)
            _orig = self.nudgePoint(rect, _orig)
            _path.appendBezierPathWithRect_(NSMakeRect(_orig.x, _orig.y, _ptSize.width, _ptSize.height))
        _color =  self.getDocument().getOption('SINGLE_POINT_COLOR')
        self.getNSColor(_color).set()
        _pathSingle.stroke()    
        _color =  self.getDocument().getOption('MULTI_POINT_COLOR')
        self.getNSColor(_color).set()
        _pathMulti.stroke()
               
    def _drawObjects(self, objs, rect, active):
        #
        # set default color if inactive
        #
        if not active:
            _color = self.getDocument().getOption('INACTIVE_LAYER_COLOR')
            self.getNSColor(_color).set()
        (_xmin, _ymin, _xmax, _ymax) = self.rectToCoordTransform(rect)

        #
        # iterate through items & draw
        #
        for _obj in objs:
            if (isinstance(_obj, Generic.graphicobject.GraphicObject) or
                isinstance(_obj, Generic.graphicobject.old_GraphicObject)):
                #
                # setup path
                #
                _path = NSBezierPath.bezierPath()
                #
                # line type
                #
                _dashList = _obj.getLinetype().getList()
                if _dashList is None:
                    _dashList = [1]
                _scaledList = []
                for _num in _dashList:
                    _scaledList.append(_num*self.__x_scale)
                _path.setLineDash_count_phase_(_scaledList, len(_scaledList), 0.0)
                #
                # line color
                #
                if active:
                    _color = _obj.getColor()
                    self.getNSColor(_color).set()
                #
                # line thickness, cap, join
                #
                _path.setLineWidth_(_obj.getThickness())
                _path.setLineJoinStyle_(NSMiterLineJoinStyle)
                _path.setLineCapStyle_(NSRoundLineCapStyle)
                #
                # items
                #
                if isinstance(_obj, Generic.segment.Segment):
                    _coords = _obj.clipToRegion(_xmin, _ymin, _xmax, _ymax)
                    if _coords is not None:
                        _x1, _y1, _x2, _y2 = _coords
                        _tp1 = self.nudgePoint(rect, NSMakePoint(_x1, _y1))
                        _tp2 = self.nudgePoint(rect, NSMakePoint(_x2, _y2))
                        _path.moveToPoint_(_tp1)
                        _path.lineToPoint_(_tp2)
                        _path.stroke()
                elif isinstance(_obj, Generic.arc.Arc):
                    _x, _y = _obj.getCenter().getCoords()
                    _point = self.nudgePoint(rect, NSMakePoint(_x, _y))
                    _r = _obj.getRadius()
                    _sa = _obj.getStartAngle()
                    _ea = _obj.getEndAngle()
                    _path.appendBezierPathWithArcWithCenter_radius_startAngle_endAngle_clockwise_(_point, _r, _sa, _ea, False)
                    _path.stroke()
                elif isinstance(_obj, Generic.circle.Circle):
                    _x, _y = _obj.getCenter().getCoords()
                    _radius = _obj.getRadius()
                    _tp = self.nudgePoint(rect, NSMakePoint(_x - _radius, _y - _radius))
                    _rect = NSMakeRect(_tp.x, _tp.y, _radius*2.0, _radius*2.0)
                    _path.appendBezierPathWithOvalInRect_(_rect)
                    _path.stroke()
                elif isinstance(_obj, Generic.segjoint.Chamfer):
                    _p1, _p2 = _obj.getMovingPoints()
                    _tp1 = self.nudgePoint(rect, NSMakePoint(_p1.x, _p1.y))
                    _tp2 = self.nudgePoint(rect, NSMakePoint(_p2.x, _p2.y))
                    _path.moveToPoint_(_tp1)
                    _path.lineToPoint_(_tp2)
                    _path.stroke()                  
                elif isinstance(_obj, Generic.segjoint.Fillet):
                    _x, _y = _obj.getCenter()
                    _p = self.nudgePoint(rect, NSMakePoint(_x, _y))
                    _r = _obj.getRadius()
                    _sa, _ea = _obj.getAngles()
                    _path.appendBezierPathWithArcWithCenter_radius_startAngle_endAngle_clockwise_(_p, _r, _sa, _ea, False)
                    _path.stroke()
                  
                elif isinstance(_obj, Generic.leader.Leader):
                    _p1, _p2, _p3 = _obj.getPoints()
                    _tp1 = self.nudgePoint(rect, NSMakePoint(_p1.x, _p1.y))
                    _tp2 = self.nudgePoint(rect, NSMakePoint(_p2.x, _p2.y))
                    _tp3 = self.nudgePoint(rect, NSMakePoint(_p3.x, _p3.y))
                    _path.moveToPoint_(_tp1)
                    _path.lineToPoint_(_tp2)
                    _path.lineToPoint_(_tp3)
                    _path.stroke()
                    _path.removeAllPoints()
                    _apts = _obj.getArrowPoints()
                    _ap1 = self.nudgePoint(rect, NSMakePoint(_apts[0], _apts[1]))
                    _ap2 = self.nudgePoint(rect, NSMakePoint(_apts[2], _apts[3]))
                    _path.moveToPoint_(_tp3)
                    _path.lineToPoint_(_ap1)
                    _path.lineToPoint_(_ap2)
                    _path.closePath()
                    _path.stroke()
                    _path.fill()
                    
                elif isinstance(_obj, Generic.polyline.Polyline):
                    for _pt in _obj.getPoints():
                        _x, _y = _pt.getCoords()
                        _point = self.nudgePoint(rect, NSMakePoint(_x, _y))
                        try:
                            _path.lineToPoint_(_point)
                        except: # catches first point
                            _path.moveToPoint_(_point)
                    _path.stroke()
            elif isinstance(_obj, Generic.text.TextBlock):
                self._drawText(_obj, rect)
            elif isinstance(_obj, Generic.dimension.HorizontalDimension):
                self._drawLinearDimension(_obj, rect, active)
            elif isinstance(_obj, Generic.dimension.VerticalDimension):
                self._drawLinearDimension(_obj, rect, active)
            elif isinstance(_obj, Generic.dimension.LinearDimension):
                self._drawLinearDimension(_obj, rect, active)
            elif isinstance(_obj, Generic.dimension.RadialDimension):
                self._drawRadialDimension(_obj, rect, active)
            elif isinstance(_obj, Generic.dimension.AngularDimension):
                self._drawAngularDimension(_obj, rect, active)
            else:
                pass
                       
    def _drawLinearDimension(self, dim, rect, active):
        _path = NSBezierPath.bezierPath()
        _path.setLineWidth_(0)
        _path.setLineJoinStyle_(NSMiterLineJoinStyle)
        _path.setLineCapStyle_(NSButtLineCapStyle)
        if active:
            _color = dim.getColor()
            self.getNSColor(_color).set()
        if dim.isModified():
            dim.calcDimValues()
            dim.reset()
        _bar1, _bar2 = dim.getDimBars()
        _crossbar = dim.getDimCrossbar()
        #
        # draw dimension lines
        #
        _ep1, _ep2 = _bar1.getEndpoints()
        _tp1 = self.nudgePoint(rect, NSMakePoint(_ep1[0], _ep1[1]))
        _tp2 = self.nudgePoint(rect, NSMakePoint(_ep2[0], _ep2[1]))
        _path.moveToPoint_(_tp1)
        _path.lineToPoint_(_tp2)
        _ep1, _ep2 = _bar2.getEndpoints()
        _tp1 = self.nudgePoint(rect, NSMakePoint(_ep1[0], _ep1[1]))
        _tp2 = self.nudgePoint(rect, NSMakePoint(_ep2[0], _ep2[1]))
        _path.moveToPoint_(_tp1)
        _path.lineToPoint_(_tp2)
        _ep1, _ep2 = _crossbar.getEndpoints()
        _tp1 = self.nudgePoint(rect, NSMakePoint(_ep1[0], _ep1[1]))
        _tp2 = self.nudgePoint(rect, NSMakePoint(_ep2[0], _ep2[1]))
        _path.moveToPoint_(_tp1)
        _path.lineToPoint_(_tp2)
        _path.stroke()
        #
        # draw endpoints
        #
        _mpts = _crossbar.getCrossbarPoints()
        _mp = _mpts[0]
        _mp1 = NSMakePoint(_mp[0], _mp[1])
        _mp = _mpts[1]
        _mp2 = NSMakePoint(_mp[0], _mp[1])
        _etype = dim.getEndpointType()
        if (_etype == Generic.dimension.Dimension.ARROW or
            _etype == Generic.dimension.Dimension.FILLED_ARROW or
            _etype == Generic.dimension.Dimension.SLASH):
            _epts = _crossbar.getMarkerPoints()
            assert len(_epts) == 4, "Bad marker point count: %d" % len(_epts)
            self._drawLineEndpoints(_etype, _epts[0:2], _mp1)
            self._drawLineEndpoints(_etype, _epts[2:4], _mp2)
                    
        elif _etype == Generic.dimension.Dimension.CIRCLE:
            _size = dim.getEndpointSize()
            self._drawCircleEndpoints(_mp1, _size)
            self._drawCircleEndpoints(_mp2, _size)
            
        else:
            pass
            
        self._drawDimensionText(dim)
            
                
    def _drawRadialDimension(self, dim, rect, active):
        _path = NSBezierPath.bezierPath()
        _path.setLineWidth_(0)
        _path.setLineJoinStyle_(NSMiterLineJoinStyle)
        _path.setLineCapStyle_(NSButtLineCapStyle)
        if active:
            _color = dim.getColor()
            self.getNSColor(_color).set()
        if dim.isModified():
            dim.calcDimValues()
            dim.reset()
        #
        # draw line
        #
        _crossbar = dim.getDimCrossbar()
        _ep1, _ep2 = _crossbar.getEndpoints()
        _tp1 = self.nudgePoint(rect, NSMakePoint(_ep1[0], _ep1[1]))
        _tp2 = self.nudgePoint(rect, NSMakePoint(_ep2[0], _ep2[1]))
        _path.moveToPoint_(_tp1)
        _path.lineToPoint_(_tp2)
        _path.stroke()
        #
        # draw endpoints
        #
        _p1, _p2 = _crossbar.getCrossbarPoints()
        _mp1 = NSMakePoint(_p1[0], _p1[1])
        _mp2 = NSMakePoint(_p2[0], _p2[1])
        _etype = dim.getEndpointType()
        _dia_mode = dim.getDiaMode()
        if (_etype == Generic.dimension.Dimension.ARROW or
            _etype == Generic.dimension.Dimension.FILLED_ARROW or
            _etype == Generic.dimension.Dimension.SLASH):
            _epts = _crossbar.getMarkerPoints()
            assert len(_epts) == 4, "Unexpect endpoint length: %d" % len(_epts)
            if _dia_mode:
                self._drawLineEndpoints(_etype, _epts[0:2], _mp1)
            self._drawLineEndpoints(_etype, _epts[2:4], _mp2)

        elif _etype == Generic.dimension.Dimension.CIRCLE:
            _size = dim.getEndpointSize()
            if _dia_mode:
                self._drawCircleEndpoints(_mp1, _size)
            self._drawCircleEndpoints(_mp2, _size)
        
        else:
            pass 

        self._drawDimensionText(dim)
        
            
    def _drawAngularDimension(self, dim, rect, active):
        _path = NSBezierPath.bezierPath()
        _path.setLineWidth_(0)
        _path.setLineJoinStyle_(NSMiterLineJoinStyle)
        _path.setLineCapStyle_(NSButtLineCapStyle)
        if active:
            _color = dim.getColor()
            self.getNSColor(_color).set()
        if dim.isModified():
            dim.calcDimValues()
            dim.reset()
        _bar1, _bar2 = dim.getDimBars()
        _crossarc = dim.getDimCrossarc()
        #
        # draw dimension lines
        #
        _ep1, _ep2 = _bar1.getEndpoints()
        _tp1 = self.nudgePoint(rect, NSMakePoint(_ep1[0], _ep1[1]))
        _tp2 = self.nudgePoint(rect, NSMakePoint(_ep2[0], _ep2[1]))
        _path.moveToPoint_(_tp1)
        _path.lineToPoint_(_tp2)
        _ep1, _ep2 = _bar2.getEndpoints()
        _tp1 = self.nudgePoint(rect, NSMakePoint(_ep1[0], _ep1[1]))
        _tp2 = self.nudgePoint(rect, NSMakePoint(_ep2[0], _ep2[1]))
        _path.moveToPoint_(_tp1)
        _path.lineToPoint_(_tp2)
        _path.stroke()
        _path.removeAllPoints()
        #
        # draw dimension arc
        #
        _vx, _vy = dim.getVertexPoint().getCoords()
        _sa = _crossarc.getStartAngle()
        _ea = _crossarc.getEndAngle()
        _rad = _crossarc.getRadius()
        _point = self.nudgePoint(rect, NSMakePoint(_vx, _vy))
        _path.appendBezierPathWithArcWithCenter_radius_startAngle_endAngle_clockwise_(_point, _rad, _sa, _ea, False)
        _path.stroke()
        #
        # draw endpoints
        #
        _p1, _p2 = _crossarc.getCrossbarPoints()
        _mp1 = NSMakePoint(_p1[0], _p1[1])
        _mp2 = NSMakePoint(_p2[0], _p2[1])
        _etype = dim.getEndpointType()
        if (_etype == Generic.dimension.Dimension.ARROW or
            _etype == Generic.dimension.Dimension.FILLED_ARROW or
            _etype == Generic.dimension.Dimension.SLASH):
            _epts = _crossarc.getMarkerPoints()
            assert len(_epts) == 4, "Unexpect endpoint length: %d" % len(_epts)
            self._drawLineEndpoints(_etype, _epts[0:2], _mp1)
            self._drawLineEndpoints(_etype, _epts[2:4], _mp2)
        elif _etype == Generic.dimension.Dimension.CIRCLE:
            _size = dim.getEndpointSize()
            self._drawCircleEndpoints(_mp1, _size)
            self._drawCircleEndpoints(_mp2, _size)
        else:
            pass 

        self._drawDimensionText(dim)
                                     
           
    def _drawCircleEndpoints(self, mp, size):         
            _path = NSBezierPath.bezierPath()
            _path.setLineWidth_(0)
            _path.setLineJoinStyle_(NSMiterLineJoinStyle)
            _path.setLineCapStyle_(NSButtLineCapStyle)
            _rad = size * 0.5
            _rect = NSMakeRect(mp.x - _rad , mp.y - _rad, size, size)
            _path.appendBezierPathWithOvalInRect_(_rect)
            _path.stroke()
            _path.fill()
      
    def _drawLineEndpoints(self, etype, ep, mp):
            _path = NSBezierPath.bezierPath()
            _path.setLineWidth_(0)
            _path.setLineJoinStyle_(NSMiterLineJoinStyle)
            _path.setLineCapStyle_(NSButtLineCapStyle)                
            if etype == Generic.dimension.Dimension.ARROW:
                    for _i in range(2):
                        _ep = ep[_i]
                        _tep = NSMakePoint(_ep[0], _ep[1])
                        _path.moveToPoint_(_tep)
                        _path.lineToPoint_(mp)
            elif etype == Generic.dimension.Dimension.FILLED_ARROW:
                    _ep = ep[0]
                    _tep = NSMakePoint(_ep[0], _ep[1])
                    _path.moveToPoint_(_tep)
                    _ep = ep[1]
                    _tep = NSMakePoint(_ep[0], _ep[1])
                    _path.lineToPoint_(_tep)
                    _path.lineToPoint_(mp)
                    _path.closePath()
            elif etype == Generic.dimension.Dimension.SLASH:
                    _ep = ep[0]
                    _tep = NSMakePoint(_ep[0], _ep[1])
                    _path.moveToPoint_(_tep)
                    _ep = ep[1]
                    _tep = NSMakePoint(_ep[0], _ep[1])
                    _path.lineToPoint_(_tep)               
            else:
                pass
            _path.stroke()
            _path.fill()
                            
    
    def _drawText(self, obj, rect):
        _gc = NSGraphicsContext.currentContext()
        _gc.setShouldAntialias_(True)
        _attrs = self.formatDictionary(obj, False)
        _x, _y = obj.getLocation()
        _point = self.nudgePoint(rect, NSMakePoint(_x, _y))
        _font = _attrs[NSFontAttributeName]
        _lineHeight = _font.defaultLineHeightForFont()
        _text = obj.getText().splitlines()
        for _str in _text:
            _as = NSAttributedString.alloc().initWithString_attributes_(_str, _attrs)
            _as.drawAtPoint_(_point)
            _point.y = _point.y - _lineHeight
        _gc.setShouldAntialias_(False)

    def _drawDimensionText(self, dim):
        _gc = NSGraphicsContext.currentContext()
        _gc.setShouldAntialias_(True)
        _x, _y = dim.getLocation()
        _dlen = dim.calculate()
        _hlen = self.getDocument().getImage().scaleLength(_dlen)
        _dims = dim.getDimensions(_hlen)
        _pds = dim.getPrimaryDimstring()
        _pdsAttr = self.formatDictionary(_pds, True)
        _dim1 = NSAttributedString.alloc().initWithString_attributes_(_dims[0], _pdsAttr)
        _dim1Size = _dim1.size()
        _dual_mode = dim.getDualDimMode()
        if _dual_mode:
            _sds = dim.getSecondaryDimstring()
            _sdsAttr = self.formatDictionary(_sds, True)
            _dim2 = NSAttributedString.alloc().initWithString_attributes_(_dims[1], _sdsAttr)
            _dim2Size = _dim2.size()
            _width = max(_dim1Size.width, _dim2Size.width) * 0.5
            # sep line & fill
            _gc.saveGraphicsState()
            _ngc = NSGraphicsContext.currentContext()
            _color = self.getDocument().getOption('BACKGROUND_COLOR')
            self.getNSColor(_color).set()
            NSRectFill(NSMakeRect((_x - _width), (_y - 2), _width * 2.0, 5))
            _ngc.restoreGraphicsState()
            _path = NSBezierPath.bezierPath()
            _path.setLineWidth_(0)
            _point = NSMakePoint((_x - _width), _y)
            _path.moveToPoint_(_point)
            _point = NSMakePoint((_x + _width), _y)
            _path.lineToPoint_(_point)
            _path.stroke()
            # prim dim
            _point = NSMakePoint((_x - _dim1Size.width * 0.5), (_y + 2)) # a little padding on the height
            _dim1.drawAtPoint_(_point)   
            # sec dim
            _point = NSMakePoint((_x - _dim2Size.width * 0.5), (_y - _dim2Size.height - 2))
            _dim2.drawAtPoint_(_point)
        else:
            _point = NSMakePoint((_x - _dim1Size.width*0.5), (_y - _dim1Size.height* 0.5))
            _dim1.drawAtPoint_(_point)
        _gc.setShouldAntialias_(False)
                 
    def pointSize(self):
        _defsize = NSMakeSize(10, 10)
        _size = self.getTransform().transformSize_(_defsize) 
        return _size

    
    def formatDictionary(self, style, db):
        """ returns attribute dictionary for NSAttributedString
        
formatDictionary(styleObj, drawsBackground):
        """
        _fm = NSFontManager.sharedFontManager()
        _family = style.getFamily()
        _style = style.getStyle()
        _weight = style.getWeight()
        _size = ceil(style.getSize() * self.__y_scale)
        if _size < 2:
            _size = 2
        _fontkey = str(_family)+str(_style)+str(_weight)+str(_size)
        if Generic.globals.NSFonts.has_key(_fontkey):
            _font = Generic.globals.NSFonts[_fontkey]
        else:
            _traits = 0
            if Generic.text.ITALIC == _style:
                _traits = _traits | NSItalicFontMask
            if Generic.text.LIGHT == _weight:
                _weight = 3
            if Generic.text.NORMAL == _weight:
                _weight = 5
            elif Generic.text.BOLD == _weight:
                _weight = 9
            elif Generic.text.HEAVY == _weight:
                _weight = 11
            _font = _fm.fontWithFamily_traits_weight_size_(_family, _traits, _weight, _size)
            if _font is None:
                _font = NSFont.userFixedPitchFontOfSize_(_size)
                if Generic.text.ITALIC == _style:
                    _font = _fm.convertFont_toHaveTrait_(_font, NSItalicFontMask)
                if _weight > 5:
                    _font = _fm.convertWeight_ofFont_(True, _font)
                elif _weight < 5:
                    _font = _fm.convertWeight_ofFont_(False, _font)
            Generic.globals.NSFonts[_fontkey] = _font
        _attrs = {}
        _attrs[NSFontAttributeName] = _font
        if _size < 4:
            _attrs[NSKernAttributeName] = 0.3
        _color = style.getColor()
        _attrs[NSForegroundColorAttributeName] = self.getNSColor(_color)
        if db:
            _color = self.getDocument().getOption('BACKGROUND_COLOR')
            _attrs[NSBackgroundColorAttributeName] = self.getNSColor(_color)
        return _attrs 
    
    
    def nudgePoint(self, rect, point):
        """ we give the points a nudge to get a single pixel line
        
nudgePoint(point)

Returns an NSPoint 
        """
        _nudge = self.__x_scale * 0.5
        _xoffset = round((point.x - rect.origin.x)/self.__x_scale)
        point.x = rect.origin.x + self.__x_scale*_xoffset + _nudge
        _nudge = self.__y_scale * 0.5
        _yoffset = round((point.y - rect.origin.y)/self.__y_scale)
        point.y = rect.origin.y + self.__y_scale*_yoffset + _nudge
        return point
       
    def translateSize(self, size):
        return self.__trans.transformSize_(size)
        
    def setTransform(self):
        _trans = NSAffineTransform.transform()
        _bounds = self.bounds()
        _frame = self.frame()
        _xs = _bounds.size.width / _frame.size.width
        _ys = _bounds.size.height / _frame.size.height
#        _xt = _bounds.origin.x * _xs
#        _yt = _bounds.origin.y * _ys
#        _trans.translateXBy_yBy_(_xt, _yt)
        _trans.scaleXBy_yBy_(_xs, _ys)
        self.__trans = _trans
        self.__x_scale = _xs
        self.__y_scale = _ys
        
    def getTransform(self):
        if self.__trans is None:
            self.setTransform()
        return self.__trans
        
 
    def rectToCoordTransform(self, rect):
        """ Converts NSRect rect into x-y coordinates of ImageDocument 
        
rectToCoordTransform(rect)

returns (xmin, ymin, xmax, ymax)
        """
        return (NSMinX(rect), NSMinY(rect), NSMaxX(rect), NSMaxY(rect))
        
    def getNSColor(self, color):
        """Retrieves an NSColor for a Color object.

getNSColor(color)

Argument "color" must be a Color object. This method returns an 
equivalent NSColor object, allocating it if necessary
        """
        if not isinstance(color, Generic.color.Color):
            raise TypeError, "Invalid Color object: " + `color`
        if Generic.globals.NSColors.has_key(color):
            _val = Generic.globals.NSColors[color]
        else:
            (_r, _g, _b) = color.getColors()
            _val = NSColor.colorWithCalibratedRed_green_blue_alpha_(_r/255.0, _g/255.0, _b/255.0, 1.0)
            Generic.globals.NSColors[color] = _val
        return _val
        

            

        
            
        
        
       

        
        
    
        
        
            


#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# gtk dimension code
#

import pygtk
pygtk.require('2.0')
import gtk
import pango

import Generic.dimension as dimension

#
# format the dimesion text string
#

def format_dim_layout(layout, ds):
    _attlist = pango.AttrList()
    _family = pango.AttrFamily(ds.getFamily())
    _family.start_index = 0
    _family.end_index = -1
    _attlist.insert(_family)
    _sty = pango.AttrStyle(ds.getStyle())
    _sty.start_index = 0
    _sty.end_index = -1
    _attlist.insert(_sty)
    _wt = pango.AttrWeight(ds.getWeight())
    _wt.start_index = 0
    _wt.end_index = -1
    _attlist.insert(_wt)
    _r, _g, _b = ds.getColor().getColors()
    _rs = int(65535 * (_r/255.0))
    _gs = int(65535 * (_g/255.0))
    _bs = int(65535 * (_b/255.0))
    _fg = pango.AttrForeground(_rs, _gs, _bs)
    _fg.start_index = 0
    _fg.end_index = -1
    _attlist.insert(_fg)
    try:
        _sz = pango.AttrSize(pango.SCALE * ds.getSize()) # need to scale
    except:
        _sz = pango.AttrSize(1024 * ds.getSize()) # need to scale
    _sz.start_index = 0
    _sz.end_index = -1
    _attlist.insert(_sz)
    layout.set_attributes(_attlist)

#
# linear dimension motion-notify handler
#

def dim_pts_motion_notify_cb(gtkimage, widget, event, tool):
    _tx, _ty = tool.getLocation()
    _px, _py = gtkimage.coordToPixTransform(_tx, _ty)
    _gc = gtkimage.getGC()
    _x = int(event.x)
    _y = int(event.y)
    _cp = tool.getCurrentPoint()
    _segs = []
    if _cp is not None:
        _xc, _yc = _cp
        _segs.append((_px, _py, _xc, _yc))
    tool.setCurrentPoint(_x, _y)
    _segs.append((_px, _py, _x, _y))
    widget.window.draw_segments(_gc, _segs)

def dim_txt_motion_notify_cb(gtkimage, widget, event, tool):
    _ix, _iy = gtkimage.getPoint()
    _gc = gtkimage.getGC()
    _ex = int(event.x)
    _ey = int(event.y)
    _cp = tool.getCurrentPoint()
    _dim = tool.getDimension()
    _bar1, _bar2 = _dim.getDimBars()
    _crossbar = _dim.getDimCrossbar()
    _segs = []
    if _cp is not None:
        _ep1, _ep2 = _bar1.getEndpoints()
        _px1, _py1 = gtkimage.coordToPixTransform(_ep1[0], _ep1[1])
        _px2, _py2 = gtkimage.coordToPixTransform(_ep2[0], _ep2[1])
        _segs.append((_px1, _py1, _px2, _py2))
        _ep1, _ep2 = _bar2.getEndpoints()
        _px1, _py1 = gtkimage.coordToPixTransform(_ep1[0], _ep1[1])
        _px2, _py2 = gtkimage.coordToPixTransform(_ep2[0], _ep2[1])
        _segs.append((_px1, _py1, _px2, _py2))
        _ep1, _ep2 = _crossbar.getEndpoints()
        _px1, _py1 = gtkimage.coordToPixTransform(_ep1[0], _ep1[1])
        _px2, _py2 = gtkimage.coordToPixTransform(_ep2[0], _ep2[1])
        _segs.append((_px1, _py1, _px2, _py2))
    tool.setCurrentPoint(_ex, _ey)
    _dim.setLocation(_ix, _iy)
    _dim.calcDimValues(False)
    _ep1, _ep2 = _bar1.getEndpoints()
    _px1, _py1 = gtkimage.coordToPixTransform(_ep1[0], _ep1[1])
    _px2, _py2 = gtkimage.coordToPixTransform(_ep2[0], _ep2[1])
    _segs.append((_px1, _py1, _px2, _py2))
    _ep1, _ep2 = _bar2.getEndpoints()
    _px1, _py1 = gtkimage.coordToPixTransform(_ep1[0], _ep1[1])
    _px2, _py2 = gtkimage.coordToPixTransform(_ep2[0], _ep2[1])
    _segs.append((_px1, _py1, _px2, _py2))
    _ep1, _ep2 = _crossbar.getEndpoints()
    _px1, _py1 = gtkimage.coordToPixTransform(_ep1[0], _ep1[1])
    _px2, _py2 = gtkimage.coordToPixTransform(_ep2[0], _ep2[1])
    _segs.append((_px1, _py1, _px2, _py2))
    widget.window.draw_segments(_gc, _segs)

#
# draw linear, horizontal, and vertical dimensions
#

def draw_linear_dimension(gtkimage, dim, active):
    _gc = gtkimage.getGC()
    _gc.set_line_attributes(1, gtk.gdk.LINE_SOLID, gtk.gdk.CAP_BUTT,
                            gtk.gdk.JOIN_MITER)
    _gc.set_function(gtk.gdk.COPY)
    _da = gtkimage.getDA()
    _window = _da.window
    _ctx = _da.get_pango_context()
    _width, _height = gtkimage.getSize()
    _pixmap = gtkimage.getPixmap()
    # dim.update()
    if dim.isModified():
        dim.calcDimValues()
        dim.reset()
    _bar1, _bar2 = dim.getDimBars()
    _crossbar = dim.getDimCrossbar()
    if active:
        _dim_color = dim.getColor()
    else:
        _dim_color = gtkimage.getOption('INACTIVE_LAYER_COLOR')
    if gtkimage.hasGTKColor(_dim_color):
        _color = gtkimage.getGTKColor(_dim_color)
    else:
        _cstring = str(_dim_color)
        _color = _da.get_colormap().alloc_color(_cstring)
        gtkimage.saveGTKColor(_dim_color, _color)
    _gc.set_foreground(_color)
    #
    # draw dimension lines
    #
    _segs = []
    _ep1, _ep2 = _bar1.getEndpoints()
    _px1, _py1 = gtkimage.coordToPixTransform(_ep1[0], _ep1[1])
    _px2, _py2 = gtkimage.coordToPixTransform(_ep2[0], _ep2[1])
    _segs.append((_px1, _py1, _px2, _py2))
    _ep1, _ep2 = _bar2.getEndpoints()
    _px1, _py1 = gtkimage.coordToPixTransform(_ep1[0], _ep1[1])
    _px2, _py2 = gtkimage.coordToPixTransform(_ep2[0], _ep2[1])
    _segs.append((_px1, _py1, _px2, _py2))
    _ep1, _ep2 = _crossbar.getEndpoints()
    _px1, _py1 = gtkimage.coordToPixTransform(_ep1[0], _ep1[1])
    _px2, _py2 = gtkimage.coordToPixTransform(_ep2[0], _ep2[1])
    _segs.append((_px1, _py1, _px2, _py2))
    #
    # draw endpoints
    #
    _etype = dim.getEndpointType()
    _mpts = _crossbar.getCrossbarPoints()
    _mp = _mpts[0]
    _mp1x, _mp1y = gtkimage.coordToPixTransform(_mp[0], _mp[1])
    _mp = _mpts[1]
    _mp2x, _mp2y = gtkimage.coordToPixTransform(_mp[0], _mp[1])
    if (_etype == dimension.Dimension.ARROW or
        _etype == dimension.Dimension.FILLED_ARROW or
        _etype == dimension.Dimension.SLASH):
        _epts = _crossbar.getMarkerPoints()
        assert len(_epts) == 4, "Bad marker point count: %d" % len(_epts)
        if _etype == dimension.Dimension.ARROW:
            _ep = _epts[0]
            _px1, _py1 = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _segs.append((_px1, _py1, _mp1x, _mp1y))
            _ep = _epts[1]
            _px1, _py1 = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _segs.append((_px1, _py1, _mp1x, _mp1y))
            _ep = _epts[2]
            _px1, _py1 = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _segs.append((_px1, _py1, _mp2x, _mp2y))
            _ep = _epts[3]
            _px1, _py1 = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _segs.append((_px1, _py1, _mp2x, _mp2y))
        elif _etype == dimension.Dimension.FILLED_ARROW:
            _tuples = []
            _ep = _epts[0]
            _px1, _py1 = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _tuples.append((_px1, _py1))
            _ep = _epts[1]
            _px1, _py1 = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _tuples.append((_px1, _py1))
            _tuples.append((_mp1x, _mp1y))
            _pixmap.draw_polygon(_gc, gtk.TRUE, _tuples)
            del _tuples[:]
            _ep = _epts[2]
            _px1, _py1 = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _tuples.append((_px1, _py1))
            _ep = _epts[3]
            _px1, _py1 = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _tuples.append((_px1, _py1))
            _tuples.append((_mp2x, _mp2y))
            _pixmap.draw_polygon(_gc, gtk.TRUE, _tuples)
        elif _etype == dimension.Dimension.SLASH:
            _ep = _epts[0]
            _px1, _py1 = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _ep = _epts[1]
            _px2, _py2 = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _segs.append((_px1, _py1, _px2, _py2))
            _ep = _epts[2]
            _px1, _py1 = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _ep = _epts[3]
            _px2, _py2 = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _segs.append((_px1, _py1, _px2, _py2))
    elif _etype == dimension.Dimension.CIRCLE:
        _upp = gtkimage.getUnitsPerPixel()
        _size = dim.getEndpointSize()/2.0
        _width = int(_size/_upp)
        _cw = _ch = _width * 2
        _xmin = _mp1x - _width
        _ymin = _mp1y - _width
        _pixmap.draw_arc(_gc, gtk.TRUE, _xmin, _ymin, _cw, _ch,
                         0, (360 * 64))
        _xmin = _mp2x - _width
        _ymin = _mp2y - _width
        _pixmap.draw_arc(_gc, gtk.TRUE, _xmin, _ymin, _cw, _ch,
                         0, (360 * 64))
    else:
        pass # catchall for now 
    #
    # draw the dimension bars, crossbars, and endpoints
    #
    _pixmap.draw_segments(_gc, _segs)
    #
    # draw dimension text
    #
    _dx, _dy = dim.getLocation()
    _pdx, _pdy = gtkimage.coordToPixTransform(_dx, _dy)    
    _dlen = dim.calculate()
    _hlen = gtkimage.scaleLength(_dlen)
    _dims = dim.getDimensions(_hlen)
    _dlines = []
    _pri_layout = pango.Layout(_ctx)
    try:
        _pri_layout.set_text(_dims[0])
    except:
        _pri_layout.set_text(_dims[0], len(_dims[0]))
    _ds = dim.getPrimaryDimstring()
    format_dim_layout(_pri_layout, _ds)
    _pri_w, _pri_h = _pri_layout.get_pixel_size()
    _layout_width = _pri_w + 2
    _layout_height = _pri_h + 2
    _dual_mode = dim.getDualDimMode()
    if _dual_mode:
        _sec_layout = pango.Layout(_ctx)
        try:
            _sec_layout.set_text(_dims[1])
        except:
            _sec_layout.set_text(_dims[1], len(_dims[1]))
        _sds = dim.getSecondaryDimstring()
        format_dim_layout(_sec_layout, _sds)
        _sec_w, _sec_h = _sec_layout.get_pixel_size()
        _layout_width = max(_pri_w, _sec_w) + 2 # extra for padding space
        _layout_height = _pri_h + _sec_h + 2 # same
    _pos = dim.getPosition() # fixme
    _hw = _layout_width//2
    _hh = _layout_height//2
    #
    # fill over the area where the text will be drawn
    #
    _box_gc = _da.get_style().fg_gc[gtk.STATE_NORMAL]
    _pixmap.draw_rectangle(_box_gc, gtk.TRUE, (_pdx - _hw), (_pdy - _hh),
                           _layout_width, _layout_height)
    #
    # draw the dimension text
    #
    _pgc = _da.get_style().black_gc
    if _dual_mode:
        _pixmap.draw_layout(_pgc, (_pdx - (_pri_w//2)), (_pdy - _pri_h - 1),
                            _pri_layout)
        _pixmap.draw_layout(_pgc, (_pdx - (_sec_w//2)), (_pdy + 1),
                            _sec_layout)
        _pixmap.draw_line(_gc, (_pdx - _hw + 1), _pdy, (_pdx + _hw - 1), _pdy)
    else:
        _pixmap.draw_layout(_pgc, (_pdx - (_pri_w//2)), (_pdy - (_pri_h//2)),
                            _pri_layout)

def add_dimension(gtkimage, tool):
    _init_func = tool.getHandler("initialize")
    tool.create(gtkimage)
    gtkimage.redraw()
    _init_func(tool)

def linear_text_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _ldim = tool.getDimension()
    _ldim.setLocation(_x, _y)
    _ldim.calcDimValues()
    _ldim.reset()
    add_dimension(gtkimage, tool)
    gtkimage.setPrompt("Click on the first point for the dimension.")
    
def linear_second_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _layers = [gtkimage.getTopLayer()]
    while len(_layers):
        _layer = _layers.pop()
        if _layer.isVisible():
            _pt = _layer.find('point', _x, _y, _tol)
            if _pt is not None:
                _x, _y = _pt.getCoords()
                tool.setSecondPoint(_layer, _pt)
                tool.setDimPosition(_x, _y)
                tool.clearCurrentPoint()
                tool.makeDimension(gtkimage)
                tool.setHandler("button_press", linear_text_button_press_cb)
                tool.setHandler("motion_notify", dim_txt_motion_notify_cb)
                gtkimage.setPrompt("Click where the dimension text should go.")
                break
        _layers.extend(_layer.getSublayers())

def linear_first_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _layers = [gtkimage.getTopLayer()]
    while len(_layers):
        _layer = _layers.pop()
        if _layer.isVisible():
            _pt = _layer.find('point', _x, _y, _tol)
            if _pt is not None:
                _x, _y = _pt.getCoords()
                tool.setLocation(_x, _y)
                tool.setFirstPoint(_layer, _pt)
                tool.setHandler("button_press", linear_second_button_press_cb)
                tool.setHandler("motion_notify", dim_pts_motion_notify_cb)
                gtkimage.setPrompt("Click on the second point for the dimension.")
                gtkimage.getGC().set_function(gtk.gdk.INVERT)
                break
        _layers.extend(_layer.getSublayers())

#
# linear dimensions
#

def linear_mode_init(tool):
    tool.setHandler("button_press", linear_first_button_press_cb)
    tool.setHandler("initialize", linear_mode_init)

#
# horizontal dimensions
#

def horizontal_mode_init(tool):
    tool.setHandler("button_press", linear_first_button_press_cb)
    tool.setHandler("initialize", horizontal_mode_init)

#
# vertical dimensions
#


def vertical_mode_init(tool):
    tool.setHandler("button_press", linear_first_button_press_cb)
    tool.setHandler("initialize", vertical_mode_init)

#
# radial dimensions
#

def draw_radial_dimension(gtkimage, rdim, active):
    _gc = gtkimage.getGC()
    _gc.set_line_attributes(1, gtk.gdk.LINE_SOLID, gtk.gdk.CAP_BUTT,
                            gtk.gdk.JOIN_MITER)
    _gc.set_function(gtk.gdk.COPY)
    _da = gtkimage.getDA()
    _window = _da.window
    _ctx = _da.get_pango_context()
    _width, _height = gtkimage.getSize()
    _pixmap = gtkimage.getPixmap()
    if rdim.isModified():
        rdim.calcDimValues()
        rdim.reset()
    _crossbar = rdim.getDimCrossbar()
    _ep1, _ep2 = _crossbar.getEndpoints()
    _p1x, _p1y = gtkimage.coordToPixTransform(_ep1[0], _ep1[1])
    _p2x, _p2y = gtkimage.coordToPixTransform(_ep2[0], _ep2[1])
    _segs = []
    _segs.append((_p1x, _p1y, _p2x, _p2y))
    _dx, _dy = rdim.getLocation()
    _pdx, _pdy = gtkimage.coordToPixTransform(_dx, _dy)
    if active:
        _dim_color = rdim.getColor()
    else:
        _dim_color = gtkimage.getOption('INACTIVE_LAYER_COLOR')
    if gtkimage.hasGTKColor(_dim_color):
        _color = gtkimage.getGTKColor(_dim_color)
    else:
        _cstring = str(_dim_color)
        _color = _da.get_colormap().alloc_color(_cstring)
        gtkimage.saveGTKColor(_dim_color, _color)
    _gc.set_foreground(_color)
    #
    # draw marker points
    #
    _mp1, _mp2 = _crossbar.getCrossbarPoints()
    _mp1x, _mp1y = gtkimage.coordToPixTransform(_mp1[0], _mp1[1])
    _etype = rdim.getEndpointType()
    _dia_mode = rdim.getDiaMode()
    if (_etype == dimension.Dimension.ARROW or
        _etype == dimension.Dimension.FILLED_ARROW or
        _etype == dimension.Dimension.SLASH):
        _epts = _crossbar.getMarkerPoints()
        assert len(_epts) == 4, "Unexpect endpoint length: %d" % len(_epts)
        if _etype == dimension.Dimension.ARROW:
            if _dia_mode:
                _ex, _ey = _epts[0]
                _epx1, _epy1 = gtkimage.coordToPixTransform(_ex, _ey)
                _segs.append((_epx1, _epy1, _mp1x, _mp1y))
                _ex, _ey = _epts[1]
                _epx2, _epy2 = gtkimage.coordToPixTransform(_ex, _ey)
                _segs.append((_epx2, _epy2, _mp1x, _mp1y))
            _mp2x, _mp2y = gtkimage.coordToPixTransform(_mp2[0], _mp2[1])
            _ex, _ey = _epts[2]
            _epx1, _epy1 = gtkimage.coordToPixTransform(_ex, _ey)
            _segs.append((_epx1, _epy1, _mp2x, _mp2y))
            _ex, _ey = _epts[3]
            _epx2, _epy2 = gtkimage.coordToPixTransform(_ex, _ey)
            _segs.append((_epx2, _epy2, _mp2x, _mp2y))
        elif _etype == dimension.Dimension.FILLED_ARROW:
            _tuples = []
            if _dia_mode:            
                _ex, _ey = _epts[0]
                _epx1, _epy1 =  gtkimage.coordToPixTransform(_ex, _ey)
                _tuples.append((_epx1, _epy1))
                _ex, _ey = _epts[1]
                _epx2, _epy2 =  gtkimage.coordToPixTransform(_ex, _ey)
                _tuples.append((_epx2, _epy2))
                _tuples.append((_mp1x, _mp1y))
                _pixmap.draw_polygon(_gc, gtk.TRUE, _tuples)
                del _tuples[:]
            _mp2x, _mp2y = gtkimage.coordToPixTransform(_mp2[0], _mp2[1])
            _ex, _ey = _epts[2]
            _epx1, _epy1 =  gtkimage.coordToPixTransform(_ex, _ey)
            _tuples.append((_epx1, _epy1))
            _ex, _ey = _epts[3]
            _epx2, _epy2 =  gtkimage.coordToPixTransform(_ex, _ey)
            _tuples.append((_epx2, _epy2))
            _tuples.append((_mp2x, _mp2y))
            _pixmap.draw_polygon(_gc, gtk.TRUE, _tuples)
        elif _etype == dimension.Dimension.SLASH:
            if _dia_mode:
                _ex, _ey = _epts[0]
                _epx1, _epy1 =  gtkimage.coordToPixTransform(_ex, _ey)
                _ex, _ey = _epts[1]
                _epx2, _epy2 =  gtkimage.coordToPixTransform(_ex, _ey)
                _segs.append((_epx1, _epy1, _epx2, _epy2))
            _ex, _ey = _epts[2]
            _epx1, _epy1 =  gtkimage.coordToPixTransform(_ex, _ey)
            _ex, _ey = _epts[3]
            _epx2, _epy2 =  gtkimage.coordToPixTransform(_ex, _ey)
            _segs.append((_epx1, _epy1, _epx2, _epy2))
    elif _etype == dimension.Dimension.CIRCLE:
        _upp = gtkimage.getUnitsPerPixel()
        _size = rdim.getEndpointSize()/2.0
        _width = int(_size/_upp)
        _cw = _ch = _width * 2
        if _dia_mode:        
            _xmin = _mp1x - _width
            _ymin = _mp1y - _width
            _pixmap.draw_arc(_gc, gtk.TRUE, _xmin, _ymin, _cw, _ch,
                             0, (360 * 64))
        _mp2x, _mp2y = gtkimage.coordToPixTransform(_mp2[0], _mp2[1])
        _xmin = _mp2x - _width
        _ymin = _mp2y - _width
        _pixmap.draw_arc(_gc, gtk.TRUE, _xmin, _ymin, _cw, _ch,
                         0, (360 * 64))
    else:
        pass # catchall, for now ...
    #
    # draw dimension bar and any markers
    #
    _pixmap.draw_segments(_gc, _segs)
    #
    # draw dimension text
    #
    _rlen = gtkimage.scaleLength(rdim.calculate())
    _dims = rdim.getDimensions(_rlen)
    _pri_layout = pango.Layout(_ctx)
    try:
        _pri_layout.set_text(_dims[0])
    except:
        _pri_layout.set_text(_dims[0], len(_dims[0]))
    _pds = rdim.getPrimaryDimstring()
    format_dim_layout(_pri_layout, _pds)
    _pri_w, _pri_h = _pri_layout.get_pixel_size()
    _layout_width = _pri_w + 2
    _layout_height = _pri_h + 2
    _dual_mode = rdim.getDualDimMode()
    if _dual_mode:
        _sec_layout = pango.Layout(_ctx)
        try:
            _sec_layout.set_text(_dims[1])
        except:
            _sec_layout.set_text(_dims[1], len(_dims[1]))
        _sds = rdim.getSecondaryDimstring()
        format_dim_layout(_sec_layout, _sds)
        _sec_w, _sec_h = _sec_layout.get_pixel_size()
        _layout_width = max(_pri_w, _sec_w) + 2 # extra for padding space
        _layout_height = _pri_h + _sec_h + 2 # same
    _pos = rdim.getPosition() # fixme
    _hw = _layout_width//2
    _hh = _layout_height//2
    #
    # fill over the area where the text will be drawn
    #
    _box_gc = _da.get_style().fg_gc[gtk.STATE_NORMAL]
    _pixmap.draw_rectangle(_box_gc, gtk.TRUE, (_pdx - _hw), (_pdy - _hh),
                           _layout_width, _layout_height)
    #
    # draw the dimension text
    #
    _pgc = _da.get_style().black_gc
    if _dual_mode:
        _pixmap.draw_layout(_pgc, (_pdx - (_pri_w//2)),
                            (_pdy - _pri_h - 1), _pri_layout)
        _pixmap.draw_layout(_pgc, (_pdx - (_sec_w//2)),
                            (_pdy + 1), _sec_layout)
        _pixmap.draw_line(_gc, (_pdx - _hw), _pdy,
                          (_pdx + _hw), _pdy)
    else:
        _pixmap.draw_layout(_pgc, (_pdx - (_pri_w//2)),
                            (_pdy - (_pri_h//2)), _pri_layout)

def radial_txt_motion_notify_cb(gtkimage, widget, event, tool):
    _gc = gtkimage.getGC()
    _x = int(event.x)
    _y = int(event.y)
    _rdim = tool.getDimension()
    _crossbar = _rdim.getDimCrossbar()
    _cp = tool.getCurrentPoint()
    _segs = []
    if _cp is not None:
        _p1, _p2 = _crossbar.getEndpoints()
        _p0x, _p0y = gtkimage.coordToPixTransform(_p1[0], _p1[1])
        _p1x, _p1y = gtkimage.coordToPixTransform(_p2[0], _p2[1])
        _segs.append((_p0x, _p0y, _p1x, _p1y))
    tool.setCurrentPoint(_x, _y)
    _ix, _iy = gtkimage.getPoint()
    _rdim.setLocation(_ix, _iy)
    _rdim.calcDimValues(False)
    _p1, _p2 = _crossbar.getEndpoints()
    _p0x, _p0y = gtkimage.coordToPixTransform(_p1[0], _p1[1])
    _p1x, _p1y = gtkimage.coordToPixTransform(_p2[0], _p2[1])
    _segs.append((_p0x, _p0y, _p1x, _p1y))
    widget.window.draw_segments(_gc, _segs)

def radial_text_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _rdim = tool.getDimension()
    _rdim.setLocation(_x, _y)
    _rdim.calcDimValues()
    _rdim.reset()
    add_dimension(gtkimage, tool)
    gtkimage.setPrompt("Click on the circle or arc to be dimensioned.")

def radial_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _dc = None
    _dl = None
    _layers = [gtkimage.getTopLayer()]
    while(len(_layers)):
        _layer = _layers.pop()
        if _layer.isVisible():
            _cobjs = (_layer.getLayerEntities("circle") +
                      _layer.getLayerEntities("arc"))
            for _cobj in _cobjs:
                _mp = _cobj.mapCoords(_x, _y, _tol)
                if _mp is not None:
                    _dc = _cobj
                    _dl = _layer
                    break
        _layers.extend(_layer.getSublayers())
    if _dc is not None:
        _x, _y = _mp
        tool.setDimObject(_dl, _dc)
        tool.setDimPosition(_x, _y)
        tool.makeDimension(gtkimage)
        tool.setHandler("motion_notify", radial_txt_motion_notify_cb)
        tool.setHandler("button_press", radial_text_button_press_cb)
        gtkimage.setPrompt("Click where the dimension text should be placed.")
        gtkimage.getGC().set_function(gtk.gdk.INVERT)

def radial_mode_init(tool):
    tool.setHandler("initialize", radial_mode_init)
    tool.setHandler("button_press", radial_button_press_cb)

#
# angular dimensions
#

def adim_txt_motion_notify_cb(gtkimage, widget, event, tool):
    _ix, _iy = gtkimage.getPoint()
    _gc = gtkimage.getGC()
    _ex = int(event.x)
    _ey = int(event.y)
    _cp = tool.getCurrentPoint()
    _adim = tool.getDimension()
    _vx, _vy = _adim.getVertexPoint().getCoords()
    _px, _py = gtkimage.coordToPixTransform(_vx, _vy)
    _win = widget.window
    _bar1, _bar2 = _adim.getDimBars()
    _crossarc = _adim.getDimCrossarc()
    _segs = []
    if _cp is not None:
        #
        # draw bars
        #
        _ep1, _ep2 = _bar1.getEndpoints()        
        _p1x, _p1y = gtkimage.coordToPixTransform(_ep1[0], _ep1[1])
        _p2x, _p2y = gtkimage.coordToPixTransform(_ep2[0], _ep2[1])
        _segs.append((_p1x, _p1y, _p2x, _p2y))
        _ep1, _ep2 = _bar2.getEndpoints()
        _p1x, _p1y = gtkimage.coordToPixTransform(_ep1[0], _ep1[1])
        _p2x, _p2y = gtkimage.coordToPixTransform(_ep2[0], _ep2[1])
        _segs.append((_p1x, _p1y, _p2x, _p2y))
        _win.draw_segments(_gc, _segs)
        del _segs[:]
        #
        # draw arc
        #
        _sa = _crossarc.getStartAngle()
        _ea = _crossarc.getEndAngle()
        _rad = int(_crossarc.getRadius()/gtkimage.getUnitsPerPixel())
        _pxmin = _px - _rad
        _pymin = _py - _rad
        _cw = _ch = _rad * 2
        if _sa < _ea:
            _sweep = _ea - _sa
        else:
            _sweep = 360.0 - (_sa - _ea)
        _win.draw_arc(_gc, gtk.FALSE, _pxmin, _pymin, _cw, _ch,
                      int(_sa * 64), int(_sweep * 64))
    tool.setCurrentPoint(_ex, _ey)
    _adim.setLocation(_ix, _iy)
    _adim.calcDimValues(False)
    #
    # draw bars
    #
    _ep1, _ep2 = _bar1.getEndpoints()        
    _p1x, _p1y = gtkimage.coordToPixTransform(_ep1[0], _ep1[1])
    _p2x, _p2y = gtkimage.coordToPixTransform(_ep2[0], _ep2[1])
    _segs.append((_p1x, _p1y, _p2x, _p2y))
    _ep1, _ep2 = _bar2.getEndpoints()
    _p1x, _p1y = gtkimage.coordToPixTransform(_ep1[0], _ep1[1])
    _p2x, _p2y = gtkimage.coordToPixTransform(_ep2[0], _ep2[1])
    _segs.append((_p1x, _p1y, _p2x, _p2y))
    _win.draw_segments(_gc, _segs)
    #
    # draw arc
    #
    _sa = _crossarc.getStartAngle()
    _ea = _crossarc.getEndAngle()
    _rad = int(_crossarc.getRadius()/gtkimage.getUnitsPerPixel())
    _pxmin = _px - _rad
    _pymin = _py - _rad
    _cw = _ch = _rad * 2
    if _sa < _ea:
        _sweep = _ea - _sa
    else:
        _sweep = 360.0 - (_sa - _ea)
    _win.draw_arc(_gc, gtk.FALSE, _pxmin, _pymin, _cw, _ch,
                  int(_sa * 64), int(_sweep * 64))

def draw_angular_dimension(gtkimage, adim, active):
    _gc = gtkimage.getGC()
    _gc.set_line_attributes(1, gtk.gdk.LINE_SOLID, gtk.gdk.CAP_BUTT,
                            gtk.gdk.JOIN_MITER)
    _gc.set_function(gtk.gdk.COPY)
    _da = gtkimage.getDA()
    _window = _da.window
    _ctx = _da.get_pango_context()
    _width, _height = gtkimage.getSize()
    _pixmap = gtkimage.getPixmap()
    # adim.update()
    if adim.isModified():
        adim.calcDimValues()
        adim.reset()
    _bar1, _bar2 = adim.getDimBars()
    _crossarc = adim.getDimCrossarc()
    if active:
        _dim_color = adim.getColor()
    else:
        _dim_color = gtkimage.getOption('INACTIVE_LAYER_COLOR')
    if gtkimage.hasGTKColor(_dim_color):
        _color = gtkimage.getGTKColor(_dim_color)
    else:
        _cstring = str(_dim_color)
        _color = _da.get_colormap().alloc_color(_cstring)
        gtkimage.saveGTKColor(_dim_color, _color)
    _gc.set_foreground(_color)
    #
    # draw dimension lines
    #
    _segs = []
    _ep1, _ep2 = _bar1.getEndpoints()        
    _p1x, _p1y = gtkimage.coordToPixTransform(_ep1[0], _ep1[1])
    _p2x, _p2y = gtkimage.coordToPixTransform(_ep2[0], _ep2[1])
    _segs.append((_p1x, _p1y, _p2x, _p2y))
    _ep1, _ep2 = _bar2.getEndpoints()
    _p1x, _p1y = gtkimage.coordToPixTransform(_ep1[0], _ep1[1])
    _p2x, _p2y = gtkimage.coordToPixTransform(_ep2[0], _ep2[1])
    _segs.append((_p1x, _p1y, _p2x, _p2y))
    #
    # draw the dimension crossbar/arc
    #
    _vx, _vy = adim.getVertexPoint().getCoords()
    _px, _py = gtkimage.coordToPixTransform(_vx, _vy)
    _dx, _dy = adim.getLocation()
    _pdx, _pdy = gtkimage.coordToPixTransform(_dx, _dy)
    _rad = int(_crossarc.getRadius()/gtkimage.getUnitsPerPixel())
    _pxmin = _px - _rad
    _pymin = _py - _rad
    _cw = _ch = _rad * 2
    _sa = _crossarc.getStartAngle()
    _ea = _crossarc.getEndAngle()
    if _sa < _ea:
        _sweep = _ea - _sa
    else:
        _sweep = 360.0 - (_sa - _ea)
    _pixmap.draw_arc(_gc, gtk.FALSE, _pxmin, _pymin, _cw, _ch,
                     int(_sa * 64), int(_sweep * 64))
    #
    # draw endpont markers
    #
    _etype = adim.getEndpointType()
    _mp1, _mp2 = _crossarc.getCrossbarPoints()
    _mp1x, _mp1y = gtkimage.coordToPixTransform(_mp1[0], _mp1[1])
    _mp2x, _mp2y = gtkimage.coordToPixTransform(_mp2[0], _mp2[1])
    if (_etype == dimension.Dimension.ARROW or
        _etype == dimension.Dimension.FILLED_ARROW or
        _etype == dimension.Dimension.SLASH):
        _epts = _crossarc.getMarkerPoints()
        assert len(_epts) == 4, "Unexpected endpoint array length: %d" % len(_epts)        
        if _etype == dimension.Dimension.ARROW:
            _ep = _epts[0]
            _p1x, _p1y = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _segs.append((_p1x, _p1y, _mp1x, _mp1y))
            _ep = _epts[1]
            _p1x, _p1y = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _segs.append((_p1x, _p1y, _mp1x, _mp1y))
            _ep = _epts[2]
            _p1x, _p1y = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _segs.append((_p1x, _p1y, _mp2x, _mp2y))
            _ep = _epts[3]
            _p1x, _p1y = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _segs.append((_p1x, _p1y, _mp2x, _mp2y))
        elif _etype == dimension.Dimension.FILLED_ARROW:
            _tuples = []
            _ep = _epts[0]
            _p1x, _p1y = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _tuples.append((_p1x, _p1y))
            _ep = _epts[1]
            _p1x, _p1y = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _tuples.append((_p1x, _p1y))
            _tuples.append((_mp1x, _mp1y))
            _pixmap.draw_polygon(_gc, gtk.TRUE, _tuples)
            del _tuples[:]
            _ep = _epts[2]
            _p1x, _p1y = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _tuples.append((_p1x, _p1y))
            _ep = _epts[3]
            _p1x, _p1y = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _tuples.append((_p1x, _p1y))
            _tuples.append((_mp2x, _mp2y))
            _pixmap.draw_polygon(_gc, gtk.TRUE, _tuples)
        elif _etype == dimension.Dimension.SLASH:
            _ep = _epts[0]
            _p1x, _p1y = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _ep = _epts[1]
            _p2x, _p2y = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _segs.append((_p1x, _p1y, _p2x, _p2y))
            _ep = _epts[2]
            _p1x, _p1y = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _ep = _epts[3]
            _p2x, _p2y = gtkimage.coordToPixTransform(_ep[0], _ep[1])
            _segs.append((_p1x, _p1y, _p2x, _p2y))
    elif _etype == dimension.Dimension.CIRCLE:
        _upp = gtkimage.getUnitsPerPixel()
        _size = adim.getEndpointSize()/2.0
        _width = int(_size/_upp)
        _cw = _ch = _width * 2
        _xmin = _mp1x - _width
        _ymin = _mp1y - _width
        _pixmap.draw_arc(_gc, gtk.TRUE, _xmin, _ymin, _cw, _ch,
                         0, (360 * 64))
        _xmin = _mp2x - _width
        _ymin = _mp2y - _width
        _pixmap.draw_arc(_gc, gtk.TRUE, _xmin, _ymin, _cw, _ch,
                         0, (360 * 64))
    else:
        pass # catchall for now ...
    #
    # draw the dimension bars and any dimension markers
    #
    _pixmap.draw_segments(_gc, _segs)    
    #
    # draw dimension text
    #
    _dlen = adim.calculate()
    _dims = adim.getDimensions(_dlen)
    _dlines = []
    _pri_layout = pango.Layout(_ctx)
    try:
        _pri_layout.set_text(_dims[0])
    except:
        _pri_layout.set_text(_dims[0], len(_dims[0]))
    _ds = adim.getPrimaryDimstring()
    format_dim_layout(_pri_layout, _ds)
    _pri_w, _pri_h = _pri_layout.get_pixel_size()
    _layout_width = _pri_w + 2
    _layout_height = _pri_h + 2
    _dual_mode = adim.getDualDimMode()
    if _dual_mode:
        _sec_layout = pango.Layout(_ctx)
        try:
            _sec_layout.set_text(_dims[1])
        except:
            _sec_layout.set_text(_dims[1], len(_dims[1]))
        _sds = adim.getSecondaryDimstring()
        format_dim_layout(_sec_layout, _sds)
        _sec_w, _sec_h = _sec_layout.get_pixel_size()
        _layout_width = max(_pri_w, _sec_w) + 2 # extra for padding space
        _layout_height = _pri_h + _sec_h + 2 # same
    _pos = adim.getPosition() # fixme
    _hw = _layout_width//2
    _hh = _layout_height//2
    #
    # fill over the area where the text will be drawn
    #
    _box_gc = _da.get_style().fg_gc[gtk.STATE_NORMAL]
    _pixmap.draw_rectangle(_box_gc, gtk.TRUE, (_pdx - _hw), (_pdy - _hh),
                           _layout_width, _layout_height)
    #
    # draw the dimension text
    #
    _pgc = _da.get_style().black_gc
    if _dual_mode:
        _pixmap.draw_layout(_pgc, (_pdx - (_pri_w//2)), (_pdy - _pri_h - 1), _pri_layout)
        _pixmap.draw_layout(_pgc, (_pdx - (_sec_w//2)), (_pdy + 1), _sec_layout)
        _pixmap.draw_line(_gc, (_pdx - _hw), _pdy, (_pdx + _hw), _pdy)
    else:
        _pixmap.draw_layout(_pgc, (_pdx - (_pri_w//2)), (_pdy - (_pri_h//2)), _pri_layout)

def _test_layer_arcs(gtkimage, lyr, x, y, tol):
    _adim = None
    for _arc in lyr.getLayerEntities("arc"):
        _arc_pt = _arc.mapCoords(x, y, tol)
        if _arc_pt is not None:
            _x, _y = _arc_pt.getCoords()
            _cp = _arc.getCenter()
            _ep1, _ep2 = _arc.getEndpoints()
            _ex, _ey = _ep1
            _p1 = lyr.find('point', _ex, _ey)
            assert _p1 is not None, "Missing arc endpoint"
            _ex, _ey = _ep2
            _p2 = lyr.find('point', _ex, _ey)
            assert _p2 is not None, "Missing arc endpoint"
            _ds = gtkimage.getOption("DIM_STYLE")
            _adim = dimension.AngularDimension(lyr, _cp, lyr, _p1, lyr, _p2,
                                               _x, _y, _ds)
            break
    return _adim

def angular_pts_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _active_layer = gtkimage.getActiveLayer()
    _pt_layer = _active_layer
    _pt = _active_layer.find('point',_x, _y, _tol)
    if _pt is None:
        _layers = [gtkimage.getTopLayer()]
        while(len(_layers)):
            _layer = _layers.pop()
            if _layer is not _active_layer and _layer.isVisible():
                _pt = _layer.find('point', _x, _y, _tol)
                if _pt is not None:
                    _pt_layer = _layer
                    break
            _layers.extend(_layer.getSublayers())
    if _pt is not None:
        _x, _y = _pt.getCoords()
        tool.storeCoords(_x, _y)
        if len(tool) == 2:
            tool.pushObject(_pt_layer)
            tool.pushObject(_pt)
        else:
            _p1 = tool.popObject()
            _l1 = tool.popObject()
            _vp = tool.popObject()
            _vl = tool.popObject()
            _ds = gtkimage.getOption("DIM_STYLE")
            _adim = dimension.AngularDimension(_vl, _vp, _l1, _p1,
                                               _pt_layer, _pt,
                                               _x, _y, _ds)
            tool.pushObject(_adim)

def angular_text_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _adim = tool.getDimension()
    _adim.setLocation(_x, _y)
    _adim.calcDimValues()
    _adim.reset()
    add_dimension(gtkimage, tool)
    gtkimage.setPrompt("Click on the angle vertex point or an arc.")

def angular_second_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _layers = [gtkimage.getTopLayer()]
    while len(_layers):
        _layer = _layers.pop()
        if _layer.isVisible():
            _pt = _layer.find('point', _x, _y, _tol)
            if _pt is not None:
                _x, _y = _pt.getCoords()
                tool.setLocation(_x, _y)
                tool.setSecondPoint(_layer, _pt)
                tool.setDimPosition(_x, _y)
                tool.makeDimension(gtkimage)
                tool.setHandler("button_press", angular_text_button_press_cb)
                tool.setHandler("motion_notify", adim_txt_motion_notify_cb)
                gtkimage.getGC().set_function(gtk.gdk.INVERT)            
                gtkimage.setPrompt("Click where the dimension text should be located.")
                break
        _layers.extend(_layer.getSublayers())
    
def angular_first_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _layers = [gtkimage.getTopLayer()]
    while len(_layers):
        _layer = _layers.pop()
        if _layer.isVisible():
            _pt = _layer.find('point', _x, _y, _tol)
            if _pt is not None:
                _x, _y = _pt.getCoords()
                tool.setLocation(_x, _y)
                tool.setFirstPoint(_layer, _pt)
                tool.setHandler("button_press", angular_second_button_press_cb)
                gtkimage.setPrompt("Click on the second point for the dimension.")
                break
        _layers.extend(_layer.getSublayers())

def _test_layer(layer, x, y, tol):
    _arc = None
    _pt = layer.find('point', x, y)
    if _pt is None:
        _pt = layer.find('point', x, y, tol)
        if _pt is None:
            _arc_pt = None
            for _arc in layer.getLayerEntities("arc"):
                _arc_pt = _arc.mapCoords(x, y, tol)
                if _arc_pt is not None:
                    break
            if _arc_pt is None:
                _arc = None # no hits on any arcs ...
    return _pt, _arc

def angular_initial_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _active_layer = gtkimage.getActiveLayer()
    _layer = _active_layer
    _pt, _arc = _test_layer(_active_layer, _x, _y, _tol)
    if _pt is None and _arc is None:
        _layers = [gtkimage.getTopLayer()]
        while len(_layers):
            _layer = _layers.pop()
            if _layer is not _active_layer and _layer.isVisible():
                _pt, _arc = _test_layer(_layer, _x, _y, _tol)
                if _pt is not None or _arc is not None:
                    break
            _layers.extend(_layer.getSublayers())
    if _pt is not None:
        tool.setVertexPoint(_layer, _pt)
        tool.setHandler("button_press", angular_first_button_press_cb)
        gtkimage.setPrompt("Click on the first endpoint for the dimension.")
    elif _arc is not None:
        _cp = _arc.getCenter()
        tool.setVertexPoint(_layer, _cp)
        _ep1, _ep2 = _arc.getEndpoints()
        _ex, _ey = _ep1
        _p1 = _layer.find('point', _ex, _ey)
        assert _p1 is not None, "Missing arc endpoint"
        tool.setFirstPoint(_layer, _p1)
        _ex, _ey = _ep2
        _p2 = _layer.find('point', _ex, _ey)
        assert _p2 is not None, "Missing arc endpoint"
        tool.setSecondPoint(_layer, _p2)
        tool.setDimPosition(_x, _y)
        tool.makeDimension(gtkimage)
        tool.setHandler("button_press", angular_text_button_press_cb)
        tool.setHandler("motion_notify", adim_txt_motion_notify_cb)
        gtkimage.getGC().set_function(gtk.gdk.INVERT)
        gtkimage.setPrompt("Click where the dimension text should be located.")        
def angular_mode_init(tool):
    tool.initialize()
    tool.setHandler("button_press", angular_initial_button_press_cb)
    tool.setHandler("initialize", angular_mode_init)



#
# Copyright (c) 2003 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# NURBS curves


import array

class Nurb(object):
    def __init__(self, ctrlpts, knots, order):
        if not isinstance(ctrlpts, list):
            raise TypeError, "Invalid control point list: " + str(ctrlpts)
        if not isinstance(knots, list):
            raise TypeError, "Invalid knot list: " + str(knots)
        if not isinstance(order, int):
            raise TypeError, "Invalid order; " + str(order)
        if order < 2 or order > 16: # what is a good max value?
            raise ValueError, "Invalid order: %d" % order
        _ctrlpts = []
        for _pt in ctrlpts:
            if not isinstance(_pt, tuple):
                raise TypeError, "Invalid control point: " + str(_pt)
            _len = len(_pt)
            if not (1 < _len < 4):
                raise ValueError, "Invalid tuple length: " + str(_pt)
            if _len == 2:
                _x, _y = _pt
                _w = 1.0
            else:
                _x, _y, _w = _pt
            if not isinstance(_x, float):
                _x = float(_x)
            if not isinstance(_y, float):
                _y = float(_y)
            if not isinstance(_w, float):
                _w = float(_w)
            if not (_w > 0.0):
                raise ValueError, "Invalid weight: %g" % _w
            _ctrlpts.append((_x, _y, _w))
        _knots = []
        for _knot in knots:
            if not isinstance(_knot, float):
                _knot = float(_knot)
            if (_knot < 0.0 or _knot > 1.0):
                raise ValueError, "Invalid knot value: %g" % _knot
            for _val in _knots:
                if _knot < (_val - 1e-10):
                    raise (ValueError, "Invalid decreasing knot: %g < %g" %
                           (_knot, _val))
            _knots.append(_knot)
        print "knots: " + str(_knots)
        print "ctrl: " + str(_ctrlpts)
        print "order: %d" % order
        _clen = len(_ctrlpts)
        if _clen < order:
            raise ValueError, "Order greater than number of control points."
        if len(_knots) != (_clen + order):
            raise ValueError, "Knot/Control Point/Order number error."
        self.__ctrlpts = _ctrlpts
        self.__knots = _knots
        self.__order = order

    def getControlPoints(self):
        return self.__ctrlpts[:]

    def getKnots(self):
        return self.__knots[:]

    def getOrder(self):
        return self.__order

    def calculate(self, count):
        if not isinstance(count, int):
            raise TypeError, "Invalid count: " + str(count)
        _cpts = self.__ctrlpts
        _knots = self.__knots
        _dt = 1.0/float(count)
        _p = self.__order - 1
        _pts = []
        for _c in range(count):
            _t = _c * _dt
            # print "time: %g" % _t
            _nx = _ny = _nw = 0.0
            for _i in range(len(_cpts)):
                # print "using cpt %d" % _i
                _x, _y, _w = _cpts[_i]
                _Ni = self._N(_i, _p, _t)
                _nx = _nx + (_Ni * _x)
                _ny = _ny + (_Ni * _y)
                _nw = _nw + (_Ni * _w)
            # print "nw: %.3f" % _nw
            # print "nx: %.3f" % _nx
            # print "ny: %.3f" % _ny
            if abs(_nw) > 1e-10:
                _pts.append((_nx/_nw, _ny/_nw))                
            else:
                print "zero weight: %f, %f" % (_nx, _ny)

        return _pts

    def _N(self, i, p, t):
        # print "_N() ..."
        _flag = False
        if abs(t - 1.0) < 1e-10 and False:
            _flag = True
        if _flag:
            print "i: %d" % i
            print "p: %d" % p
            print "t: %.3f" % t
        _knots = self.__knots
        _ki = _knots[i]
        _kin = _knots[i + 1]
        if _flag:
            print "ki: %.3f" % _ki
            print "kin: %.3f" % _kin
        if p == 0:
            if ((_ki - 1e-10) < t < _kin):
                _val = 1.0
            else:
                _val = 0.0
        else:
            _kip = _knots[i + p]
            _kipn = _knots[i + p + 1]
            if _flag:
                print "kip: %.3f" % _kip
                print "kipn: %.3f" % _kipn
            _t1 = 0.0
            _v1 = _kip - _ki
            if abs(_v1) > 1e-10:
                _v2 = t - _ki
                if abs(_v2) > 1e-10:
                    _t1 = (_v2/_v1) * self._N(i, (p - 1), t)
            _t2 = 0.0
            _v1 = _kipn - _kin
            if abs(_v1) > 1e-10:
                _v2 = _kipn - t
                if abs(_v2) > 1e-10:
                    _t2 = (_v2/_v1) * self._N((i + 1), (p - 1), t)
            _val = _t1 + _t2
        if _flag:
            print "val: %f" % _val
        return _val

    def writedata(self, count, fname):
        if not isinstance(count, int):
            raise TypeError, "Invalid count: " + str(count)
        _f = file(fname, "w")
        for _pt in self.calculate(count):
            _x, _y = _pt
            _f.write("%f %f\n" % (_x, _y))
        _f.close()
        _f = file('control_points', "w")
        for _pt in self.getControlPoints():
            _x, _y, _w = _pt # ignore weight
            _f.write("%f %f\n" % (_x, _y))
        _f.close()
        _f = file('knots', "w")
        for _knot in self.getKnots():
            _f.write("%f 0.0\n" % _knot)
        _f.close()

    def _NN(self, i, p, t):
        _cpts = self.__ctrlpts
        _cl = len(_cpts)
        _knots = self.__knots
        _kl = len(_knots) - 1
        _val = _kl * [0.0]
        #
        # calculate values for 0
        #
        for _i in range(_kl):
            if ((_knots[i] - 1e-10) < t < _knots[i + 1]):
                _val[_i] = 1.0
        #
        # calculate values up to the degree
        #
        for _j in range(1, (p + 1)):
            for _i in range(_kl - _j):
                _ki = _knots[_i]
                _kin = _knots[_i + 1]
                _kip = _knots[_i + p]
                _kipn = _knots[_i + p + 1]
                _t1 = 0.0
                _n = _val[_i]
                if abs(_n) > 1e-10:
                    _v1 = _kip - _ki
                    if abs(_v1) > 1e-10:
                        _v2 = t - _ki
                        if abs(_v2) > 1e-10:
                            _t1 = (_v2/_v1) * _n
                _t2 = 0.0
                _n = _val[_i + 1]
                if abs(_n) > 1e-10:
                    _v1 = _kipn - _kin
                    if abs(_v1) > 1e-10:
                        _v2 = _kipn - t
                        if abs(_v2) > 1e-10:
                            _t2 = (_v2/_v1) * _n
                _val[_i] = _t1 + _t2
